# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import warnings
import requests

try:
    import ujson

    requests.models.complexjson = ujson
except ImportError as exc_:
    warnings.warn(
        "usjon is not installed. You could potentially get huge benefit if installing it"
    )

import time
import numpy as np
from tqdm import tqdm

from grid2op.MakeEnv import make

try:
    from lightsim2grid import LightSimBackend

    bkclass = LightSimBackend
    # raise ImportError()
except ImportError as exc_:
    from grid2op.Backend import PandaPowerBackend

    bkclass = PandaPowerBackend
    warnings.warn(
        "lightsim2grid is not installed. You could potentially get huge benefit if installing it"
    )
    pass

URL = " http://127.0.0.1:5000"
env_name = "l2rpn_case14_sandbox"
env_name = "l2rpn_neurips_2020_track1_small"
env_name = "l2rpn_neurips_2020_track2_small"


def assert_rec_equal(li1, li2):
    try:
        len(li1)
        assert len(li1) == len(li2), "wrong seed length"
        for (el1, el2) in zip(li1, li2):
            assert_rec_equal(el1, el2)
    except Exception as exc_:
        if isinstance(li1, np.ndarray):
            if li1.shape == ():
                li1 = []
        assert li1 == li2, "wrong seed value"


if __name__ == "__main__":
    # create the real environment
    real_env = make(env_name, backend=bkclass())
    real_obs = real_env.reset()
    client = requests.session()

    print('Test "make"')
    # test the "make" endpoint of the API
    resp_make = client.get(f"{URL}/make/{env_name}")

    # check that the creation is working
    if resp_make.status_code != 200:
        raise RuntimeError("Environment not created response not 200")
    resp_make_json = resp_make.json()
    if "id" not in resp_make_json:
        raise RuntimeError("Environment not created (due to id not in json)")
    if "obs" not in resp_make_json:
        raise RuntimeError("Environment not created (due to obs not in json)")

    if env_name == "l2rpn_case14_sandbox":
        # the other envs are stochastic so this test cannot work right now (this is why we used the "seed" just after)
        reic_obs_json = resp_make_json["obs"]
        reic_obs = copy.deepcopy(real_obs)
        reic_obs.set_game_over()
        assert reic_obs != real_obs, "resetting the observation did not work"
        reic_obs.from_json(reic_obs_json)
        are_same = reic_obs == real_obs
        diff_obs = reic_obs - real_obs
        assert are_same, "obs received and obs computed are not the same"

    # make a step with do nothing
    id_env = resp_make_json["id"]

    print('Test "seed"')
    seed_used = 0
    resp_seed = client.post(
        f"{URL}/seed/{env_name}/{id_env}",
        json={"seed": seed_used},
        # headers={"X-CSRFToken": csrf_token}
    )
    if resp_seed.status_code != 200:
        raise RuntimeError("Environment not seeded response not 200")
    resp_seed_json = resp_seed.json()
    if "seeds" not in resp_seed_json:
        raise RuntimeError("Environment not seeded (due to seeds not in json)")
    res_seed = real_env.seed(seed_used)
    res_seed = list(res_seed)
    assert_rec_equal(res_seed, resp_seed_json["seeds"])

    print('Test "reset"')
    resp_reset = client.get(f"{URL}/reset/{env_name}/{id_env}")
    if resp_reset.status_code != 200:
        raise RuntimeError("Environment not reset response not 200")
    resp_reset_json = resp_reset.json()
    if "obs" not in resp_reset_json:
        raise RuntimeError("Environment not reset (due to obs not in json)")
    real_obs = real_env.reset()
    reic_obs_json = resp_reset_json["obs"]
    reic_obs = copy.deepcopy(real_obs)
    reic_obs.set_game_over()
    assert reic_obs != real_obs, "resetting the observation did not work"
    reic_obs.from_json(reic_obs_json)
    are_same = reic_obs == real_obs
    obs_diff, attr_diff = reic_obs.where_different(real_obs)
    for el in attr_diff:
        if np.max(np.abs(getattr(obs_diff, el))) > 1e-4:
            tmp_ = np.max(np.abs(getattr(obs_diff, el)))
            import pdb

            pdb.set_trace()
            raise RuntimeError(
                f"ERROR: after reset, attribute {el} is not the same (max={tmp_:.6f})"
            )
    if not are_same:
        warnings.warn(
            "obs received and obs computed are not the exactly the same "
            "(but equal up to some small value (1e-4))"
        )

    print('Test "set_id"')
    resp_set_id = client.post(f"{URL}/set_id/{env_name}/{id_env}", json={"id": 0})
    if resp_set_id.status_code != 200:
        raise RuntimeError("set_id not working: response is not 200")
    resp_set_id_json = resp_set_id.json()
    if "info" not in resp_set_id_json:
        raise RuntimeError("set_id not working: info is not in response")

    resp_reset = client.get(f"{URL}/reset/{env_name}/{id_env}")
    if resp_seed.status_code != 200:
        raise RuntimeError("Environment not reset response not 200")
    resp_reset_json = resp_reset.json()
    if "obs" not in resp_reset_json:
        raise RuntimeError("Environment not reset (due to obs not in json)")
    real_env.set_id(0)
    real_obs = real_env.reset()
    reic_obs_json = resp_reset_json["obs"]
    reic_obs = copy.deepcopy(real_obs)
    reic_obs.set_game_over()
    assert reic_obs != real_obs, "resetting the observation did not work"
    reic_obs.from_json(reic_obs_json)
    are_same = reic_obs == real_obs
    assert are_same, "obs received and obs computed are not the same"

    print('Test "set_thermal_limit"')
    th_lim = real_env.get_thermal_limit().tolist()
    resp_set_thermal_limit = client.post(
        f"{URL}/set_thermal_limit/{env_name}/{id_env}", json={"thermal_limits": th_lim}
    )
    if resp_set_thermal_limit.status_code != 200:
        raise RuntimeError("set_thermal_limit not working: response is not 200")
    resp_set_thermal_limit_json = resp_set_thermal_limit.json()
    if "env_name" not in resp_set_thermal_limit_json:
        raise RuntimeError("set_thermal_limit not working: info is not in response")

    print('Test "get_thermal_limit"')
    resp_get_thermal_limit = client.get(f"{URL}/get_thermal_limit/{env_name}/{id_env}")
    if resp_get_thermal_limit.status_code != 200:
        raise RuntimeError("get_thermal_limit not working: response is not 200")
    resp_get_thermal_limit_json = resp_get_thermal_limit.json()
    if "thermal_limit" not in resp_get_thermal_limit_json:
        raise RuntimeError(
            "get_thermal_limit not working: thermal_limit is not in response"
        )
    assert (
        resp_get_thermal_limit_json["thermal_limit"] == th_lim
    ), "get_thermal_limit not working: wrong thermal limit"

    print('Test "get_path_env"')
    resp_get_path_env = client.get(f"{URL}/get_path_env/{env_name}/{id_env}")
    if resp_get_path_env.status_code != 200:
        raise RuntimeError("get_path_env not working: response is not 200")
    resp_get_path_env_json = resp_get_path_env.json()
    if "path" not in resp_get_path_env_json:
        raise RuntimeError("get_path_env not working: path is not in response")
    assert (
        resp_get_path_env_json["path"] == real_env.get_path_env()
    ), "get_path_env not working: wrong path"

    print('Test "fast_forward_chronics"')
    nb_step_forward = 10
    resp_fast_forward_chronics = client.post(
        f"{URL}/fast_forward_chronics/{env_name}/{id_env}",
        json={"nb_step": nb_step_forward},
    )
    if resp_fast_forward_chronics.status_code != 200:
        raise RuntimeError("set_id not working: response is not 200")
    resp_fast_forward_chronics_json = resp_fast_forward_chronics.json()
    if "env_name" not in resp_fast_forward_chronics_json:
        raise RuntimeError(
            "get_thermal_limit not working: thermal_limit is not in response"
        )
    act = real_env.action_space()
    real_env.fast_forward_chronics(nb_step_forward)
    obs, reward, done, info = real_env.step(act)
    resp_step = client.post(
        f"{URL}/step/{env_name}/{id_env}", json={"action": act.to_json()}
    )
    # check obs are equals
    reic_obs_json = resp_step.json()["obs"]
    reic_obs = copy.deepcopy(obs)
    reic_obs.set_game_over()
    assert reic_obs != real_obs, "resetting the observation did not work"
    reic_obs.from_json(reic_obs_json)
    are_same = reic_obs == obs
    assert (
        are_same
    ), "obs received and obs computed are not the same after fast forwarding"

    print('Test "step"')
    real_obs = real_env.reset()
    resp_reset = client.get(f"{URL}/reset/{env_name}/{id_env}")
    if resp_seed.status_code != 200:
        raise RuntimeError(
            "Environment not reset response not 200, fail just before assessing step"
        )

    nb_step = 0
    obs_me = copy.deepcopy(real_obs)
    while True:
        act = real_env.action_space()
        obs, reward, done, info = real_env.step(act)
        resp_step = client.post(
            f"{URL}/step/{env_name}/{id_env}", json={"action": act.to_json()}
        )
        if resp_step.status_code != 200:
            raise RuntimeError("Step not successful not 200")
        resp_step_json = resp_step.json()
        if "obs" not in resp_step_json:
            raise RuntimeError("Environment not created (due to obs not in json)")

        assert resp_step_json["done"] == done
        assert resp_step_json["reward"] == reward
        if done:
            break

        reic_obs_json = resp_step_json["obs"]
        obs_me.set_game_over()
        assert (
            obs_me != obs
        ), f"resetting the observation did not work for step {nb_step}"
        obs_me.from_json(reic_obs_json)
        is_correct = obs_me == obs
        try:
            assert (
                is_correct
            ), f"obs received and obs computed are not the same after step for step {nb_step}"
        except AssertionError as exc_:
            obs_diff, attr_diff = obs_me.where_different(obs)
            # import pdb
            # pdb.set_trace()
        nb_step += 1

    print('Test "close"')
    resp_close = client.get(f"{URL}/close/{env_name}/{id_env}")
    if resp_close.status_code != 200:
        raise RuntimeError("close not working: response is not 200")
    resp_close_json = resp_close.json()
    if "env_name" not in resp_close_json:
        raise RuntimeError("close not working: env_name is not in response")
    # TODO test all methods fails (get_thermal_limit apparently work...)

    print('Test "perfs"')
    print(f"Time on a local env: (using {bkclass.__name__})")
    env_perf = make(env_name, backend=bkclass())
    env_perf.reset()
    env_perf.seed(seed_used)
    obs = env_perf.reset()
    time_for_step = 0
    nb_step_local = 0
    with tqdm(desc="local env") as pbar:
        while True:
            act = real_env.action_space()
            beg_step = time.perf_counter()
            obs, reward, done, info = env_perf.step(act)
            time_for_step += time.perf_counter() - beg_step
            if done:
                break
            nb_step_local += 1
            pbar.update(1)

    print("Time on the remote env:")
    resp_make_perf = client.get(f"{URL}/make/{env_name}")
    id_env_perf = resp_make_perf.json()["id"]
    _ = client.post(f"{URL}/seed/{env_name}/{id_env_perf}", json={"seed": seed_used})
    _ = client.get(f"{URL}/reset/{env_name}/{id_env_perf}")
    time_for_step_api = 0.0
    time_for_all_api = 0.0
    time_convert = 0.0
    time_get_json = 0.0
    nb_step_api = 0
    with tqdm(desc="remote env") as pbar:
        while True:
            act = real_env.action_space()
            beg_step = time.perf_counter()
            act_as_json = act.to_json()
            resp_step = client.post(
                f"{URL}/step/{env_name}/{id_env_perf}", json={"action": act_as_json}
            )
            after_step = time.perf_counter()
            time_for_step_api += after_step - beg_step
            resp_step_json = resp_step.json()
            time_get_json += time.perf_counter() - after_step
            reic_obs_json = resp_step_json["obs"]
            beg_convert = time.perf_counter()
            obs.from_json(reic_obs_json)
            time_convert += time.perf_counter() - beg_convert
            time_for_all_api += time.perf_counter() - beg_step
            if resp_step_json["done"]:
                break
            pbar.update(1)
            nb_step_api += 1

    print(f"\tEnv name: {env_name} with {real_env.n_sub} substations")
    print(f"\tNumber of step for local env {nb_step_local}")
    print(f"\tNumber of step for api env {nb_step_api}")
    print(f"\tTime to compute all, for the normal env: {time_for_step:.2f}s")
    print(f"\tTime to compute all, for the api env: {time_for_all_api:.2f}s")
    print(f"\t\tTime to do the step, for the api env: {time_for_step_api:.2f}s")
    print(f"\t\tTime to get the json from the http request: {time_get_json:.2f}s")
    print(f"\t\tTime to convert from json: {time_convert:.2f}s")
    print(f"\tSpeed up (for normal env): {time_for_all_api/time_for_step:.2f}")



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.utils.l2rpn_2020_scores import ScoreL2RPN2020
from grid2op.Reward import L2RPNSandBoxScore, _NewRenewableSourcesUsageScore, _AlertTrustScore
from grid2op.utils.underlying_statistics import EpisodeStatistics
from grid2op.Exceptions import Grid2OpException


class ScoreL2RPN2023(ScoreL2RPN2020):
    """
    This class allows to compute the same score as the one computed for the L2RPN 2023 competitions.

    It uses some "EpisodeStatistics" of the environment to compute these scores. These statistics, if not available
    are computed at the initialization.

    When using it a second time these information are reused.


    This scores is the combination of the `ScoreL2RPN2020` score and some extra scores based on the assistant feature
    (alert) and the use of new renewable energy sources.

    Examples
    ---------
    This class can be used as follow:

    .. code-block:: python

        import grid2op
        from grid2op.utils import ScoreL2RPN2023
        from grid2op.Agent import DoNothingAgent

        env = grid2op.make("l2rpn_case14_sandbox")
        nb_scenario = 2
        my_score = ScoreL2RPN2023(env,
                                  nb_scenario=nb_scenario,
                                  env_seeds=[0 for _ in range(nb_scenario)],
                                  agent_seeds=[0 for _ in range(nb_scenario)]
                                  )

        my_agent = DoNothingAgent(env.action_space)
        print(my_score.get(my_agent))


    Notes
    -------
    To prevent overfitting, we strongly recommend you to use the :func:`grid2op.Environment.Environment.train_val_split`
    and use this function on the built validation set only.

    Also note than computing the statistics, and evaluating an agent on a whole dataset of multiple GB can take a
    really long time and a lot of memory. This fact, again, plea in favor of using this function only on
    a validation set.

    We also strongly recommend to set the seeds of your agent (agent_seeds)
    and of the environment (env_seeds) if you want to use this feature. Reproducibility is really important if you
    want to make progress.

    .. warning::

        The triggering (or not) of the recomputation of the statistics is not perfect for now.
        We recommend you to use always
        the same seeds (`env_seeds` and `agent_seeds` key word argument of this functions)
        and the same parameters (`env.parameters`) when using a given environments.

        You might need to clean it manually if you change
        one of theses things by calling :func:`ScoreL2RPN2020.clear_all()` function .

    """

    def __init__(
            self,
            env,
            env_seeds=None,
            agent_seeds=None,
            nb_scenario=16,
            min_losses_ratio=0.8,
            verbose=0,
            max_step=-1,
            nb_process_stats=1,
            scores_func={
                "grid_operational_cost": L2RPNSandBoxScore,
                "assistant_confidence": _AlertTrustScore,
                "new_renewable_sources_usage": _NewRenewableSourcesUsageScore,
            },
            score_names=["grid_operational_cost_scores",
                         "assistant_confidence_scores",
                         "new_renewable_sources_usage_scores"],
            add_nb_highres_sim=False,
            scale_assistant_score=100.0,
            scale_nres_score=100.,
            weight_op_score=0.6,
            weight_assistant_score=0.25,
            weight_nres_score=0.15,
            min_nres_score=-100,
            min_assistant_score=-300
    ):
        ScoreL2RPN2020.__init__(
            self,
            env=env,
            env_seeds=env_seeds,
            agent_seeds=agent_seeds,
            nb_scenario=nb_scenario,
            min_losses_ratio=min_losses_ratio,
            verbose=verbose,
            max_step=max_step,
            nb_process_stats=nb_process_stats,
            scores_func=scores_func,
            score_names=score_names,
            add_nb_highres_sim=add_nb_highres_sim,
        )
        weights=np.array([weight_op_score,weight_assistant_score,weight_nres_score])
        total_weights = weights.sum()
        if total_weights != 1.0:
            raise Grid2OpException(
                'The weights of each component of the score shall sum to 1'
            )
        if np.any(weights <0):
            raise Grid2OpException(
                'All weights should be positive'
            )

        self.scale_assistant_score = scale_assistant_score
        self.scale_nres_score = scale_nres_score
        self.weight_op_score = weight_op_score
        self.weight_assistant_score = weight_assistant_score
        self.weight_nres_score = weight_nres_score
        self.min_nres_score = min_nres_score
        self.min_assistant_score = min_assistant_score

    def _compute_episode_score(
            self,
            ep_id,  # the ID here, which is an integer and is not the ID from chronics balblabla
            meta,
            other_rewards,
            dn_metadata,
            no_ov_metadata,
            score_file_to_use="grid_operational_cost_scores",
    ):
        """
        Performs the rescaling of the score given the information stored in the "statistics" of this
        environment.

        This computes the score for a single episode. The loop to compute the score for all the
        episodes is the same as for l2rpn_2020_scores and is then reused.
        """

        # compute the operational score
        op_score, n_played, total_ts = super()._compute_episode_score(
            ep_id,
            meta,
            other_rewards,
            dn_metadata,
            no_ov_metadata,
            # score_file_to_use should match the
            # L2RPNSandBoxScore key in
            # self.scores_func
            score_file_to_use=score_file_to_use,
        )
        # should match underlying_statistics.run_env `dict_kwg["other_rewards"][XXX] = ...`
        # XXX is right now f"{EpisodeStatistics.KEY_SCORE}_{nm}" [this should match the XXX]

        # retrieve nres_score
        new_renewable_sources_usage_score_nm = "new_renewable_sources_usage_scores"
        real_nm = EpisodeStatistics._nm_score_from_attr_name(new_renewable_sources_usage_score_nm)
        key_score_file = f"{EpisodeStatistics.KEY_SCORE}_{real_nm}"
        nres_score = float(other_rewards[-1][key_score_file])
        nres_score = max(nres_score, self.min_nres_score / self.scale_nres_score)
        nres_score = self.scale_nres_score * nres_score

        # assistant_score
        assistant_confidence_score_nm = "assistant_confidence_scores"
        real_nm = EpisodeStatistics._nm_score_from_attr_name(assistant_confidence_score_nm)
        key_score_file = f"{EpisodeStatistics.KEY_SCORE}_{real_nm}"
        assistant_confidence_score = float(other_rewards[-1][key_score_file])
        assistant_confidence_score = max(assistant_confidence_score, self.min_assistant_score / self.scale_assistant_score)
        assistant_score = self.scale_assistant_score * assistant_confidence_score

        ep_score = (
                self.weight_op_score * op_score + self.weight_nres_score * nres_score + self.weight_assistant_score * assistant_score
        )
        return (ep_score, op_score, nres_score, assistant_score), n_played, total_ts



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import time

import os

import grid2op.MakeEnv.PathUtils
from grid2op.Exceptions import UnknownEnv
from grid2op.MakeEnv.UserUtils import list_available_local_env
from grid2op.MakeEnv.Make import _retrieve_github_content

_LIST_REMOTE_URL = (
    "https://api.github.com/repos/bdonnot/grid2op-datasets/contents/updates.json"
)
_LIST_REMOTE_ENV_HASH = (
    "https://api.github.com/repos/bdonnot/grid2op-datasets/contents/env_hashes.json"
)


def _write_file(path_local_env, new_config, file_name):
    with open(os.path.join(path_local_env, file_name), "w", encoding="utf-8") as f:
        f.write(new_config)


def update_env(env_name=None):
    """
    This function allows you to retrieve the latest version of the some of files used to create the
    environment.

    File can be for example "config.py" or "prod_charac.csv" or "difficulty_levels.json".

    Parameters
    ----------
    env_name: ``str``
        The name of the environment you want to update the config file (must be an environment you
        have already downloaded). If ``None`` it will look for updates for all the environments
        locally available.

    Examples
    --------
    Here is an example on how to for the update of your environments:

    .. code-block:: python

        import grid2op
        grid2op.update_env()
        # it will download the files "config.py" or "prod_charac.csv" or "difficulty_levels.json"
        # of your local environment to match the latest version available.

    """
    _update_files(env_name=env_name)


def _update_file(dict_, env_name, file_name):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Update a single file of a single environment.

    File can be for example "config.py" or "prod_charac.csv" or "difficulty_levels.json".
    """
    baseurl, filename = dict_["base_url"], dict_["filename"]
    url_ = baseurl + filename
    time.sleep(1)
    new_config = _retrieve_github_content(url_, is_json=False)
    path_local_env = os.path.join(grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA, env_name)
    if os.path.exists(os.path.join(path_local_env, ".multimix")):
        # this is a multimix env ...
        mixes = os.listdir(path_local_env)
        for mix in mixes:
            mix_dir = os.path.join(path_local_env, mix)
            if os.path.exists(os.path.join(mix_dir, file_name)):
                # this is indeed a mix
                _write_file(mix_dir, new_config, file_name=file_name)
    else:
        _write_file(path_local_env, new_config, file_name=file_name)
    print(
        '\t Successfully updated file "{}" for environment "{}"'.format(
            file_name, env_name
        )
    )


def _do_env_need_update(env_name, env_hashes):
    if env_name not in env_hashes:
        # no hash for this environment is provided, i don't know, so in doubt i need to update it (old behaviour)
        return True
    else:
        # i check if "my" hash is different from the remote hash
        base_path = grid2op.get_current_local_dir()
        hash_remote_hex = env_hashes[env_name]
        hash_local = _hash_env(os.path.join(base_path, env_name))
        hash_local_hex = hash_local.hexdigest()
        res = hash_remote_hex != hash_local_hex
        return res


def _update_files(env_name=None, answer_json=None, env_hashes=None):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Update all the "modified" files of a given environment. If ``None`` is provided as input, all local environments
    will be checked for update.

    Parameters
    ----------
    env_name: ``str``
        Name of the environment you want to update (should be locally available)

    """
    avail_envs = list_available_local_env()

    if answer_json is None:
        # optimization to retrieve only once this file
        answer_json = _retrieve_github_content(_LIST_REMOTE_URL)

    if env_hashes is None:
        # optimization to retrieve only once this file
        env_hashes = _retrieve_github_content(_LIST_REMOTE_ENV_HASH)

    if env_name is None:
        # i update all the files for all the environments
        for env_name in avail_envs:
            _update_files(env_name, answer_json=answer_json, env_hashes=env_hashes)
    else:
        # i update the files for only an environment
        if env_name in avail_envs:
            need_update = _do_env_need_update(env_name, env_hashes)
            if env_name in answer_json and need_update:
                dict_main = answer_json[env_name]
                for k, dict_ in dict_main.items():
                    _update_file(dict_, env_name, file_name=k)
            elif need_update and env_name not in answer_json:
                print(
                    f'Environment: "{env_name}" is not up to date, but we did not found any files to update. '
                    f'IF this environment is officially supported by grid2op (see full list at '
                    f'https://grid2op.readthedocs.io/en/latest/available_envs.html#description-of-some-environments) '
                    f'Please write an issue at :\n\t\t'
                    f'https://github.com/rte-france/Grid2Op/issues/new?assignees=&labels=question&title=Environment%20{env_name}%20is%20not%20up%20to%20date%20but%20I%20cannot%20update%20it.&body=%3c%21%2d%2dDescribe%20shortly%20the%20context%20%2d%2d%3e%0d'
                )
            else:
                # environment is up to date
                print('Environment "{}" is up to date'.format(env_name))
        else:
            raise UnknownEnv(
                'Impossible to locate the environment named "{}". Have you downlaoded it?'
                "".format(env_name)
            )


# TODO make that a method of the environment maybe ?
def _hash_env(
    path_local_env,
    hash_=None,
    blocksize=64,  # TODO is this correct ?
):
    import hashlib  # lazy import

    if hash_ is None:
        # we use this as it is supposedly faster than md5
        # we don't really care about the "secure" part of it (though it's a nice tool to have)
        hash_ = hashlib.blake2b()
    if os.path.exists(os.path.join(path_local_env, ".multimix")):
        # this is a multi mix, so i need to run through all sub env
        mixes = sorted(os.listdir(path_local_env))
        for mix in mixes:
            mix_dir = os.path.join(path_local_env, mix)
            if os.path.isdir(mix_dir):
                hash_ = _hash_env(mix_dir, hash_=hash_, blocksize=blocksize)
    else:
        # i am hashing a regular environment
        # first i hash the config files
        for fn_ in [
            "alerts_info.json",
            "config.py",
            "difficulty_levels.json",
            "grid.json",
            "grid_layout.json",
            "prods_charac.csv",
            "storage_units_charac.csv",
            # chronix2grid files, if any
            "loads_charac.csv",
            "params.json",
            "params_load.json",
            "params_loss.json",
            "params_opf.json",
            "params_res.json",
            "scenario_params.json",
        ]:  # list the file we want to hash (we don't hash everything
            full_path_file = os.path.join(path_local_env, fn_)
            import re

            if os.path.exists(full_path_file):
                with open(full_path_file, "r", encoding="utf-8") as f:
                    text_ = f.read()
                    text_ = re.sub(
                        "\s", "", text_
                    )  # this is done to ensure a compatibility between platform
                    # sometime git replaces the "\r\n" in windows with "\n" on linux / macos and it messes
                    # up the hash
                    hash_.update(text_.encode("utf-8"))

        # now I hash the chronics
        # but as i don't want to read every chronics (for time purposes) i will only hash the names
        # of all the chronics
        path_chronics = os.path.join(path_local_env, "chronics")
        for chron_name in sorted(os.listdir(path_chronics)):
            hash_.update(chron_name.encode("utf-8"))
    return hash_



__all__ = [
    "Grid2OpException",
    "EnvError",
    "IncorrectNumberOfLoads",
    "IncorrectNumberOfGenerators",
    "IncorrectNumberOfLines",
    "IncorrectNumberOfSubstation",
    "IncorrectNumberOfStorages",
    "IncorrectNumberOfElements",
    "IncorrectPositionOfLoads",
    "IncorrectPositionOfGenerators",
    "IncorrectPositionOfLines",
    "IncorrectPositionOfStorages",
    "UnknownEnv",
    "MultiEnvException",
    "IllegalAction",
    "OnProduction",
    "VSetpointModified",
    "ActiveSetPointAbovePmax",
    "ActiveSetPointBelowPmin",
    "OnLoad",
    "OnLines",
    "InvalidReconnection",
    "UnitCommitorRedispachingNotAvailable",
    "NotEnoughGenerators",
    "GeneratorTurnedOffTooSoon",
    "GeneratorTurnedOnTooSoon",
    "InvalidRedispatching",
    "InvalidBusStatus",
    "InvalidNumberOfObjectEnds",
    "InvalidNumberOfLines",
    "InvalidNumberOfGenerators",
    "InvalidNumberOfLoads",
    "UnrecognizedAction",
    "InvalidLineStatus",
    "InvalidStorage",
    "InvalidCurtailment",
    "AmbiguousAction",
    "NonFiniteElement",
    "AmbiguousActionRaiseAlert",
    "DivergingPowerFlow",
    "BaseObservationError",
    "NoForecastAvailable",
    "SimulateError",
    "SimulateUsedTooMuchThisStep",
    "SimulateUsedTooMuchThisEpisode",
    "ChronicsError",
    "ChronicsNotFoundError",
    "InsufficientData",
    "BackendError",
    "PlotError",
    "OpponentError",
    "UsedRunnerError",
    "NotEnoughAttentionBudget",
    "AgentError",
    "SimulatorError",
    "HandlerError"
]

from grid2op.Exceptions.Grid2OpException import Grid2OpException

from grid2op.Exceptions.EnvExceptions import EnvError
from grid2op.Exceptions.EnvExceptions import IncorrectNumberOfLoads
from grid2op.Exceptions.EnvExceptions import IncorrectNumberOfGenerators
from grid2op.Exceptions.EnvExceptions import IncorrectNumberOfLines
from grid2op.Exceptions.EnvExceptions import IncorrectNumberOfSubstation
from grid2op.Exceptions.EnvExceptions import IncorrectNumberOfStorages
from grid2op.Exceptions.EnvExceptions import IncorrectNumberOfElements
from grid2op.Exceptions.EnvExceptions import IncorrectPositionOfLoads
from grid2op.Exceptions.EnvExceptions import IncorrectPositionOfGenerators
from grid2op.Exceptions.EnvExceptions import IncorrectPositionOfLines
from grid2op.Exceptions.EnvExceptions import IncorrectPositionOfStorages
from grid2op.Exceptions.EnvExceptions import UnknownEnv
from grid2op.Exceptions.EnvExceptions import MultiEnvException

from grid2op.Exceptions.IllegalActionExceptions import IllegalAction
from grid2op.Exceptions.IllegalActionExceptions import OnProduction
from grid2op.Exceptions.IllegalActionExceptions import VSetpointModified
from grid2op.Exceptions.IllegalActionExceptions import ActiveSetPointAbovePmax
from grid2op.Exceptions.IllegalActionExceptions import ActiveSetPointBelowPmin
from grid2op.Exceptions.IllegalActionExceptions import OnLoad
from grid2op.Exceptions.IllegalActionExceptions import OnLines
from grid2op.Exceptions.IllegalActionExceptions import InvalidReconnection
from grid2op.Exceptions.IllegalActionExceptions import (
    UnitCommitorRedispachingNotAvailable,
)

from grid2op.Exceptions.AmbiguousActionExceptions import NotEnoughGenerators
from grid2op.Exceptions.AmbiguousActionExceptions import GeneratorTurnedOffTooSoon
from grid2op.Exceptions.AmbiguousActionExceptions import GeneratorTurnedOnTooSoon
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidRedispatching
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidBusStatus
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidNumberOfObjectEnds
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidNumberOfLines
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidNumberOfGenerators
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidNumberOfLoads
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidCurtailment
from grid2op.Exceptions.AmbiguousActionExceptions import UnrecognizedAction
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidLineStatus
from grid2op.Exceptions.AmbiguousActionExceptions import InvalidStorage
from grid2op.Exceptions.AmbiguousActionExceptions import AmbiguousAction
from grid2op.Exceptions.AmbiguousActionExceptions import NonFiniteElement
from grid2op.Exceptions.AmbiguousActionExceptions import AmbiguousActionRaiseAlert

from grid2op.Exceptions.PowerflowExceptions import DivergingPowerFlow

from grid2op.Exceptions.ObservationExceptions import BaseObservationError
from grid2op.Exceptions.ObservationExceptions import NoForecastAvailable
from grid2op.Exceptions.ObservationExceptions import SimulateError
from grid2op.Exceptions.ObservationExceptions import SimulateUsedTooMuchThisStep
from grid2op.Exceptions.ObservationExceptions import SimulateUsedTooMuchThisEpisode

from grid2op.Exceptions.ChronicsExceptions import ChronicsError
from grid2op.Exceptions.ChronicsExceptions import ChronicsNotFoundError
from grid2op.Exceptions.ChronicsExceptions import InsufficientData
from grid2op.Exceptions.handlers_exceptions import HandlerError

from grid2op.Exceptions.BackendExceptions import BackendError

from grid2op.Exceptions.PlotExceptions import PlotError

from grid2op.Exceptions.OpponentError import OpponentError

from grid2op.Exceptions.RunnerError import UsedRunnerError

from grid2op.Exceptions.AttentionBudgetExceptions import NotEnoughAttentionBudget

from grid2op.Exceptions.agentError import AgentError

from grid2op.Exceptions.simulatorExceptions import SimulatorError



# Copyright (c) 2019-2021, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import time
import requests
import os
import warnings
import pkg_resources
from typing import Union, Optional
import logging

from grid2op.Environment import Environment
from grid2op.MakeEnv.MakeFromPath import make_from_dataset_path, ERR_MSG_KWARGS
from grid2op.Exceptions import Grid2OpException, UnknownEnv
import grid2op.MakeEnv.PathUtils
from grid2op.MakeEnv.PathUtils import _create_path_folder
from grid2op.Download.DownloadDataset import _aux_download

_VAR_FORCE_TEST = "_GRID2OP_FORCE_TEST"

DEV_DATA_FOLDER = pkg_resources.resource_filename("grid2op", "data")
DEV_DATASET = os.path.join(DEV_DATA_FOLDER, "{}")
TEST_DEV_ENVS = {
    "blank": DEV_DATASET.format("blank"),
    "rte_case14_realistic": DEV_DATASET.format("rte_case14_realistic"),
    "rte_case14_redisp": DEV_DATASET.format("rte_case14_redisp"),
    "rte_case14_test": DEV_DATASET.format("rte_case14_test"),
    "rte_case5_example": DEV_DATASET.format("rte_case5_example"),
    "rte_case118_example": DEV_DATASET.format("rte_case118_example"),
    "rte_case14_opponent": DEV_DATASET.format("rte_case14_opponent"),
    "l2rpn_wcci_2020": DEV_DATASET.format("l2rpn_wcci_2020"),
    "l2rpn_neurips_2020_track2": DEV_DATASET.format("l2rpn_neurips_2020_track2"),
    "l2rpn_neurips_2020_track1": DEV_DATASET.format("l2rpn_neurips_2020_track1"),
    "l2rpn_case14_sandbox": DEV_DATASET.format("l2rpn_case14_sandbox"),
    "l2rpn_case14_sandbox_diff_grid": DEV_DATASET.format("l2rpn_case14_sandbox_diff_grid"),
    "l2rpn_icaps_2021": DEV_DATASET.format("l2rpn_icaps_2021"),
    "l2rpn_wcci_2022_dev": DEV_DATASET.format("l2rpn_wcci_2022_dev"),
    "l2rpn_wcci_2022": DEV_DATASET.format("l2rpn_wcci_2022_dev"),
    "l2rpn_idf_2023": DEV_DATASET.format("l2rpn_idf_2023"),
    # educational files
    "educ_case14_redisp": DEV_DATASET.format("educ_case14_redisp"),
    "educ_case14_storage": DEV_DATASET.format("educ_case14_storage"),
    # keep the old names for now
    "case14_realistic": DEV_DATASET.format("rte_case14_realistic"),
    "case14_redisp": DEV_DATASET.format("rte_case14_redisp"),
    "case14_test": DEV_DATASET.format("rte_case14_test"),
    "case5_example": DEV_DATASET.format("rte_case5_example"),
    "case14_fromfile": DEV_DATASET.format("rte_case14_test"),
}

_REQUEST_FAIL_EXHAUSTED_ERR = (
    'Impossible to retrieve data at "{}".\n'
    "If the problem persists, please contact grid2op developers by sending an issue at "
    "https://github.com/rte-france/Grid2Op/issues"
)
_REQUEST_FAIL_RETRY_ERR = (
    'Failure to get a response from the url "{}".\n'
    "Retrying... {} attempt(s) remaining"
)
_REQUEST_EXCEPT_RETRY_ERR = (
    'Exception in getting an answer from "{}".\n' "Retrying... {} attempt(s) remaining"
)

_LIST_REMOTE_URL = (
    "https://api.github.com/repos/bdonnot/grid2op-datasets/contents/datasets.json"
)
_LIST_REMOTE_KEY = "download_url"
_LIST_REMOTE_INVALID_CONTENT_JSON_ERR = (
    "Impossible to retrieve available datasets. "
    "File could not be converted to json. "
    "Parsing error:\n {}"
)
_LIST_REMOTE_CORRUPTED_CONTENT_JSON_ERR = (
    "Corrupted json retrieved from github api. "
    "Please wait a few minutes and try again. "
    "If the error persist, contact grid2op devs by making an issue at "
    "\n\thttps://github.com/rte-france/Grid2Op/issues/new/choose"
)
_LIST_REMOTE_INVALID_DATASETS_JSON_ERR = (
    "Impossible to retrieve available datasets. "
    "File could not be converted to json. "
    'The error was \n"{}"'
)

_FETCH_ENV_UNKNOWN_ERR = (
    'Impossible to find the environment named "{}".\n'
    "Current available environments are:\n{}"
)

_MULTIMIX_FILE = ".multimix"

_MAKE_DEV_ENV_WARN = (
    "You are using a development environment. "
    "This environment is not intended for training agents. It might not be up to date "
    'and its primary use if for tests (hence the "test=True" you passed as argument). '
    "Use at your own risk."
)
_MAKE_DEV_ENV_DEPRECATED_WARN = (
    'Dev env "{}" has been deprecated '
    "and will be removed in future version.\n"
    'Please update to dev envs starting by "rte" or "l2rpn"'
)
_MAKE_FIRST_TIME_WARN = (
    'It is the first time you use the environment "{}".\n'
    "We will attempt to download this environment from remote"
)
_MAKE_UNKNOWN_ENV = 'Impossible to load the environment named "{}".'

_EXTRACT_DS_NAME_CONVERT_ERR = (
    'The "dataset_name" argument '
    "should be convertible to string, "
    'but "{}" was provided.'
)
_EXTRACT_DS_NAME_RECO_ERR = (
    'Impossible to recognize the environment name from path "{}"'
)

def _force_test_dataset():
    res = False
    if _VAR_FORCE_TEST in os.environ:
        try:
            var_int = int(os.environ[_VAR_FORCE_TEST])
        except Exception as exc_:
            warnings.warn(f"The environment variable {_VAR_FORCE_TEST}, "
                          f"used to force the \"test=True\" in grid2op "
                          f"cannot be converted to an integer with error "
                          f"\"{exc_}\". As it is set nonetheless, we "
                          f"assume you want to force \"test=True\".")
            var_int = 1
        res = var_int >= 1
    return res

def _send_request_retry(url, nb_retry=10, gh_session=None):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
    """
    if nb_retry <= 0:
        raise Grid2OpException(_REQUEST_FAIL_EXHAUSTED_ERR.format(url))

    if gh_session is None:
        gh_session = requests.Session()

    try:
        response = gh_session.get(url=url)
        if response.status_code == 200:
            return response
        warnings.warn(_REQUEST_FAIL_RETRY_ERR.format(url, nb_retry - 1))
        time.sleep(1)
        return _send_request_retry(url, nb_retry=nb_retry - 1, gh_session=gh_session)
    except Grid2OpException:
        raise
    except KeyboardInterrupt:
        raise
    except Exception as exc_:
        warnings.warn(_REQUEST_EXCEPT_RETRY_ERR.format(url, nb_retry - 1))
        time.sleep(1)
        return _send_request_retry(url, nb_retry=nb_retry - 1, gh_session=gh_session)


def _retrieve_github_content(url, is_json=True):
    answer = _send_request_retry(url)
    try:
        answer_json = answer.json()
    except Exception as e:
        raise Grid2OpException(_LIST_REMOTE_INVALID_CONTENT_JSON_ERR.format(e))

    if _LIST_REMOTE_KEY not in answer_json:
        raise Grid2OpException(_LIST_REMOTE_CORRUPTED_CONTENT_JSON_ERR)
    time.sleep(1)
    avail_datasets = _send_request_retry(answer_json[_LIST_REMOTE_KEY])
    if is_json:
        try:
            res = avail_datasets.json()
        except Exception as e:
            raise Grid2OpException(_LIST_REMOTE_INVALID_DATASETS_JSON_ERR.format(e))
    else:
        res = avail_datasets.text
    return res


def _list_available_remote_env_aux():
    return _retrieve_github_content(url=_LIST_REMOTE_URL)


def _fecth_environments(dataset_name):
    avail_datasets_json = _list_available_remote_env_aux()
    if not dataset_name in avail_datasets_json:
        known_ds = sorted(avail_datasets_json.keys())
        raise UnknownEnv(_FETCH_ENV_UNKNOWN_ERR.format(dataset_name, known_ds))
    # url = _FETCH_ENV_TAR_URL.format(avail_datasets_json[dataset_name], dataset_name)
    dict_ = avail_datasets_json[dataset_name]
    baseurl, filename = dict_["base_url"], dict_["filename"]
    url = baseurl + filename
    # name is "tar.bz2" so i need to get rid of 2 extensions
    ds_name_dl = os.path.splitext(os.path.splitext(filename)[0])[0]
    return url, ds_name_dl


def _extract_ds_name(dataset_path):
    """
    If a path is provided, clean it to have a proper datasetname.

    If a dataset name is already provided, then i just returns it.

    Parameters
    ----------
    dataset_path: ``str``
        The path in the form of a

    Returns
    -------
    dataset_name: ``str``
        The name of the dataset (all lowercase, without "." etc.)

    """

    try:
        dataset_path = str(dataset_path)
    except Exception as exc_:
        raise Grid2OpException(
            _EXTRACT_DS_NAME_CONVERT_ERR.format(dataset_path)
        ) from exc_

    try:
        dataset_name = os.path.split(dataset_path)[-1]
    except Exception as exc_:
        raise UnknownEnv(_EXTRACT_DS_NAME_RECO_ERR.format(dataset_path)) from exc_
    dataset_name = dataset_name.lower().rstrip().lstrip()
    dataset_name = os.path.splitext(dataset_name)[0]
    return dataset_name


def _aux_is_multimix(dataset_path):
    if os.path.exists(os.path.join(dataset_path, _MULTIMIX_FILE)):
        return True
    return False


def _aux_make_multimix(
    dataset_path,
    test=False,
    experimental_read_from_local_dir=False,
    _add_to_name="",
    _compat_glop_version=None,
    logger=None,
    **kwargs
) -> Environment:
    # Local import to prevent imports loop
    from grid2op.Environment import MultiMixEnvironment

    return MultiMixEnvironment(
        dataset_path,
        experimental_read_from_local_dir=experimental_read_from_local_dir,
        _test=test,
        _add_to_name=_add_to_name,
        _compat_glop_version=_compat_glop_version,
        logger=logger,
        **kwargs
    )


def make(
    dataset : Union[str, os.PathLike]=None,
    *,
    test : bool=False,
    logger: Optional[logging.Logger]=None,
    experimental_read_from_local_dir : bool=False,
    _add_to_name : str="",
    _compat_glop_version : Optional[str]=None,
    **kwargs
) -> Environment:
    """
    This function is a shortcut to rapidly create some (pre defined) environments within the grid2op framework.

    Other environments, with different powergrids will be made available in the future and will be easily downloadable
    using this function.

    It mimic the `gym.make` function.

    .. versionchanged:: 1.9.3
        Remove the possibility to use this function with arguments (force kwargs)
        
    Parameters
    ----------

    dataset: ``str`` or path
        Name of the environment you want to create

    test: ``bool``
        Whether you want to use a test environment (**NOT** recommended). Use at your own risk.

    logger: 
        If you want to use a specific logger for environment and all other 
        grid2op objects, you can put it here. This feature is still under development.
        
    experimental_read_from_local_dir: ``bool``
        Grid2op "embed" the grid description into the description of the classes
        themselves. By default this is done "on the fly" (when the environment is created)
        but for some usecase (especially ones involving multiprocessing or "pickle")
        it might not be easily usable. If you encounter issues with pickle or multi
        processing, you can set this flag to ``True``. See the doc of 
        :func:`grid2op.Environment.BaseEnv.generate_classes` for more information.
        
    kwargs:
        Other keyword argument to give more control on the environment you are creating. See
        the Parameters information of the :func:`make_from_dataset_path`.

    _add_to_name:
        Internal, do not use (and can only be used when setting "test=True"). If
        `experimental_read_from_local_dir` is set to True, this has no effect.

    _compat_glop_version:
        Internal, do not use (and can only be used when setting "test=True")

    Returns
    -------
    env: :class:`grid2op.Environment.Environment`
        The created environment.

    Examples
    --------

    If you want to create the environment "rte_case14_realistic":

    .. code-block: python

        import grid2op
        env_name = "rte_case14_realistic"  # or any other supported environment
        env = grid2op.make(env_name)
        # env implements the openai gym interface (env.step, env.render, env.reset etc.)

    **NB** the first time you type this command, the dataset (approximately 300 MB for this one) will be
    downloaded from the internet, sizes vary per dataset.

    """
    if _force_test_dataset():
        if not test:
            warnings.warn(f"The environment variable \"{_VAR_FORCE_TEST}\" is defined so grid2op will be forced in \"test\" mode. "
                          f"This is equivalent to pass \"grid2op.make(..., test=True)\" and prevents any download of data.")
            test = True
    
    if dataset is None:
        raise Grid2OpException("Impossible to create an environment without its name. Please call something like: \n"
                               "> env = grid2op.make('l2rpn_case14_sandbox') \nor\n"
                               "> env = grid2op.make('rte_case14_realistic')")

    accepted_kwargs = ERR_MSG_KWARGS.keys() | {"dataset", "test"}
    for el in kwargs:
        if el not in accepted_kwargs:
            raise Grid2OpException(
                'The keyword argument "{}" you provided is invalid. Possible keyword '
                'arguments to create environments are "{}".'
                "".format(el, sorted(accepted_kwargs))
            )
    # Select how to create the environment:
    # Default with make from path
    make_from_path_fn = make_from_dataset_path
    
    # dataset arg is a valid path: load it
    if os.path.exists(dataset):
        # check if its a test environment
        if test:
            _add_to_name_tmp = _add_to_name
            _compat_glop_version_tmp = _compat_glop_version
            test_tmp = True
        else:
            _add_to_name_tmp = ""
            _compat_glop_version_tmp = None
            test_tmp = False

        # Check if multimix from path
        if _aux_is_multimix(dataset) and not test_tmp:
            make_from_path_fn = _aux_make_multimix
        elif _aux_is_multimix(dataset) and test_tmp:

            def make_from_path_fn_(*args, **kwargs):
                if not "logger" in kwargs:
                    kwargs["logger"] = logger
                if not "experimental_read_from_local_dir" in kwargs:
                    kwargs[
                        "experimental_read_from_local_dir"
                    ] = experimental_read_from_local_dir
                return _aux_make_multimix(*args, test=True, **kwargs)

            make_from_path_fn = make_from_path_fn_
        
        if not "logger" in kwargs:
            kwargs["logger"] = logger
        if not "experimental_read_from_local_dir" in kwargs:
            kwargs[
                "experimental_read_from_local_dir"
            ] = experimental_read_from_local_dir
        
        return make_from_path_fn(
            dataset_path=dataset,
            _add_to_name=_add_to_name_tmp,
            _compat_glop_version=_compat_glop_version_tmp,
            **kwargs
        )

    # Not a path: get the dataset name and cache path
    dataset_name = _extract_ds_name(dataset)
    real_ds_path = os.path.join(
        grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA, dataset_name
    )

    # Unknown dev env
    if test and dataset_name not in TEST_DEV_ENVS:
        raise Grid2OpException(_MAKE_UNKNOWN_ENV.format(dataset))
    
    # Known test env and test flag enabled
    if test:
        warnings.warn(_MAKE_DEV_ENV_WARN)
        # Warning for deprecated dev envs
        if not (
            dataset_name.startswith("rte")
            or dataset_name.startswith("l2rpn")
            or dataset_name.startswith("educ")
        ):
            warnings.warn(_MAKE_DEV_ENV_DEPRECATED_WARN.format(dataset_name))
        ds_path = TEST_DEV_ENVS[dataset_name]
        # Check if multimix from path
        if _aux_is_multimix(ds_path):

            def make_from_path_fn_(*args, **kwargs):
                if "logger" not in kwargs:
                    kwargs[
                        "logger"
                    ] = logger  # foward the logger if not present already
                return _aux_make_multimix(*args, test=True, **kwargs)

            make_from_path_fn = make_from_path_fn_

        return make_from_path_fn(
            dataset_path=ds_path,
            logger=logger,
            _add_to_name=_add_to_name,
            _compat_glop_version=_compat_glop_version,
            experimental_read_from_local_dir=experimental_read_from_local_dir,
            **kwargs
        )

    # Env directory is present in the DEFAULT_PATH_DATA
    if os.path.exists(real_ds_path):
        if _aux_is_multimix(real_ds_path):
            make_from_path_fn = _aux_make_multimix
        return make_from_path_fn(
            real_ds_path,
            logger=logger,
            experimental_read_from_local_dir=experimental_read_from_local_dir,
            **kwargs
        )

    # Env needs to be downloaded
    warnings.warn(_MAKE_FIRST_TIME_WARN.format(dataset_name))
    _create_path_folder(grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA)
    url, ds_name_dl = _fecth_environments(dataset_name)
    _aux_download(
        url, dataset_name, grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA, ds_name_dl
    )

    # Check if multimix from path
    if _aux_is_multimix(real_ds_path):
        make_from_path_fn = _aux_make_multimix
    return make_from_path_fn(
        dataset_path=real_ds_path,
        logger=logger,
        experimental_read_from_local_dir=experimental_read_from_local_dir,
        **kwargs
    )



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Exceptions.Grid2OpException import Grid2OpException


# ambiguous action
class AmbiguousAction(Grid2OpException):
    """
    This exception indicate that the :class:`grid2op.BaseAction` is ambiguous.
    It could be understood differently according
    to the backend used.

    Such a kind of action are forbidden in this package. These kind of exception are mainly thrown by the
    :class:`grid2op.BaseAction.BaseAction` in
    the :func:`grid2op.BaseAction.update` and :func:`grid2op.BaseAction.__call__` methods.

    As opposed to a :class:`IllegalAction` an :class:`AmbiguousAction` is forbidden for all the backend,
    in all the scenarios.

    It doesn't depend on the implemented rules.
    """

    pass


class InvalidLineStatus(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that the
     :class:`grid2op.BaseAction.BaseAction` is ambiguous due to powerlines manipulation.
    """

    pass


class InvalidStorage(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that the
     :class:`grid2op.BaseAction.BaseAction` is ambiguous due to storage unit manipulation.
    """

    pass


class UnrecognizedAction(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that the
    :class:`grid2op.BaseAction.BaseAction` is  ambiguous due to the bad formatting of the action.
    """

    pass


class InvalidNumberOfLoads(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction` is ambiguous because an incorrect number of loads tries to be modified.
    """

    pass


class InvalidNumberOfGenerators(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction`
    is ambiguous because an incorrect number of generator tries to be modified.
    """

    pass


class InvalidNumberOfLines(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction`
    is ambiguous because an incorrect number of lines tries to be modified.
    """

    pass


class InvalidNumberOfObjectEnds(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction`
    is ambiguous because an incorrect number of object at a substation try to be modified.
    """

    pass


class InvalidBusStatus(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction`
    try to both "set" and "switch" some bus to which an object is connected.
    """

    pass


class InvalidRedispatching(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction`
    try to apply an invalid redispatching strategy.
    """

    pass


class InvalidCurtailment(AmbiguousAction):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that
    the :class:`grid2op.BaseAction.BaseAction`
    try to apply an invalid curtailment strategy.
    """

    pass


class GeneratorTurnedOnTooSoon(InvalidRedispatching):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that a generator has been turned on
    before gen_min_up_time time steps.
    """

    pass


class GeneratorTurnedOffTooSoon(InvalidRedispatching):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that a generator has been turned off
    before gen_min_down_time time steps.
    """

    pass


class NotEnoughGenerators(InvalidRedispatching):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that there is not enough turned off
    generators to meet the demand.
    """

    pass


class NonFiniteElement(InvalidRedispatching):
    """
    This is a more precise exception than :class:`AmbiguousAction` indicating that an action / observation
    non initialized (full of Nan)
    has been loaded by the "from_vect" method.
    """

    pass

class AmbiguousActionRaiseAlert(AmbiguousAction):
    """Raise if the type of action is ambiguous due to the 'raiseAlert' part"""
    pass


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import warnings

from grid2op.Exceptions import AmbiguousAction
from grid2op.Action.baseAction import BaseAction


class VoltageOnlyAction(BaseAction):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        This is the main class used by the voltage controller.

    This class is here to serve as a base class for the controller of the voltages (if any). It allows to perform
    only modification of the generator voltage set point.

    Only action of type "injection" are supported, and only setting "prod_v" keyword.
    """

    authorized_keys = {"injection"}
    attr_list_vect = ["prod_v"]
    attr_list_set = set(attr_list_vect)
    _shunt_added = False
    _first_init = True

    def __init__(self):
        """
        See the definition of :func:`BaseAction.__init__` and of :class:`BaseAction` for more information. Nothing more is done
        in this constructor.

        """
        BaseAction.__init__(self)

        if VoltageOnlyAction._shunt_added is False and self.shunts_data_available:
            VoltageOnlyAction.attr_list_vect += ["shunt_p", "shunt_q", "shunt_bus"]
            VoltageOnlyAction.authorized_keys.add("shunt")
            VoltageOnlyAction._shunt_added = True

        if VoltageOnlyAction._first_init is True:
            self._update_value_set()
            VoltageOnlyAction._first_init = False

    def _check_dict(self):
        """
        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Check that nothing, beside prod_v has been updated with this action.

        Returns
        -------

        """
        if self._dict_inj:
            for el in self._dict_inj:
                if el not in self.attr_list_vect:
                    raise AmbiguousAction(
                        'Impossible to modify something different than "prod_v" using '
                        '"VoltageOnlyAction" action.'
                    )

    def update(self, dict_):
        """
        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        As its original implementation, this method allows modifying the way a dictionary can be mapped to a valid
        :class:`BaseAction`.

        It has only minor modifications compared to the original :func:`BaseAction.update` implementation, most notably, it
        doesn't update the :attr:`BaseAction._dict_inj`. It raises a warning if attempting to change them.

        Parameters
        ----------
        dict_: :class:`dict`
            See the help of :func:`BaseAction.update` for a detailed explanation. **NB** all the explanations concerning the
            "injection", "change bus", "set bus", or "change line status" are irrelevant for this subclass.

        Returns
        -------
        self: :class:`PowerLineSet`
            Return object itself thus allowing multiple calls to "update" to be chained.

        """
        self._reset_vect()

        if dict_ is not None:
            for kk in dict_.keys():
                if kk not in self.authorized_keys:
                    warn = 'The key "{}" used to update an action will be ignored. Valid keys are {}'
                    warn = warn.format(kk, self.authorized_keys)
                    warnings.warn(warn)

            self._digest_injection(dict_)
            self._digest_shunt(dict_)
            self._check_dict()
        return self



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Game, Grid2Game a gamified platform to interact with grid2op environments.

__all__ = [
    "BaseAgent",
    "DoNothingAgent",
    "OneChangeThenNothing",
    "GreedyAgent",
    "PowerLineSwitch",
    "TopologyGreedy",
    "AgentWithConverter",
    "RandomAgent",
    "DeltaRedispatchRandomAgent",
    "MLAgent",
    "RecoPowerlineAgent",
    "FromActionsListAgent",
    "RecoPowerlinePerArea"
]

from grid2op.Agent.baseAgent import BaseAgent
from grid2op.Agent.doNothing import DoNothingAgent
from grid2op.Agent.oneChangeThenNothing import OneChangeThenNothing
from grid2op.Agent.greedyAgent import GreedyAgent
from grid2op.Agent.powerlineSwitch import PowerLineSwitch
from grid2op.Agent.topologyGreedy import TopologyGreedy
from grid2op.Agent.agentWithConverter import AgentWithConverter
from grid2op.Agent.randomAgent import RandomAgent
from grid2op.Agent.deltaRedispatchRandomAgent import DeltaRedispatchRandomAgent
from grid2op.Agent.mlAgent import MLAgent
from grid2op.Agent.recoPowerlineAgent import RecoPowerlineAgent
from grid2op.Agent.fromActionsListAgent import FromActionsListAgent
from grid2op.Agent.recoPowerLinePerArea import RecoPowerlinePerArea



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Reward.baseReward import BaseReward

class _AssistantConfidenceScore(BaseReward):
    """

    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            It **must not** serve as a reward. This scored needs to be **MINIMIZED**,
            and a reward needs to be maximized! Also, this "reward" is not scaled or anything. Use it as your
            own risk.

    Implemented as a reward to make it easier to use in the context of the L2RPN competitions, this "reward"
    computed the assistant "confidence score", which evaluates how confident an agent was in its actions for handling
    unforeseen line l disconnection events prior to occurring.
    It should not be used to train an agent.
    
    """
    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)
        
    def initialize(self, env):
        self.reset(env)
        
    def __call__(self, env, obs, is_done):
        return 0.
        
class _AssistantCostScore(BaseReward):
    """

    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            It **must not** serve as a reward. This scored needs to be **MINIMIZED**,
            and a reward needs to be maximized! Also, this "reward" is not scaled or anything. Use it as your
            own risk.

    Implemented as a reward to make it easier to use in the context of the L2RPN competitions, this "reward"
    computed the assistant"cost score", which penalized the number of alarm the assistant have produced.
    It should not be used to train an agent.
    
    """
    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)
        
    def initialize(self, env):
        self.reset(env)
        
    def __call__(self, env, obs, is_done):
        return 0.



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class CombinedReward(BaseReward):
    """
    This class allows to combine multiple pre defined reward. The reward it computes will
    be the sum of all the sub rewards it is made of.

    Each sub reward is identified by a key.

    It is used a bit differently that the other rewards. See the section example for more information.

    Examples
    --------

    .. code-block:: python

        import grid2op
        from grid2op.Reward import GameplayReward, FlatReward, CombinedReward

        env = grid2op.make(..., reward_class=CombinedReward)
        cr = self.env.get_reward_instance()
        cr.addReward("Gameplay", GameplayReward(), 1.0)
        cr.addReward("Flat", FlatReward(), 1.0)
        cr.initialize(self.env)

        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())

        # reward here is computed by summing the results of what would have
        # given `GameplayReward` and the one from `FlatReward`

    """

    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)
        self.reward_min = dt_float(0.0)
        self.reward_max = dt_float(0.0)
        self.rewards = {}

    def addReward(self, reward_name, reward_instance, reward_weight=1.0):
        self.rewards[reward_name] = {
            "instance": reward_instance,
            "weight": dt_float(reward_weight),
        }
        return True

    def removeReward(self, reward_name):
        if reward_name in self.rewards:
            self.rewards.pop(reward_name)
            return True
        return False

    def updateRewardWeight(self, reward_name, reward_weight):
        if reward_name in self.rewards:
            self.rewards[reward_name]["weight"] = reward_weight
            return True
        return False

    def __iter__(self):
        for k, v in super().__iter__():
            yield (k, v)
        for k, v in self.rewards.items():
            r_dict = dict(v["instance"])
            r_dict["weight"] = float(v["weight"])
            yield (k, r_dict)

    def initialize(self, env):
        self.reward_min = dt_float(0.0)
        self.reward_max = dt_float(0.0)

        for key, reward in self.rewards.items():
            reward_w = reward["weight"]
            reward_instance = reward["instance"]
            reward_instance.initialize(env)
            self.reward_max += dt_float(reward_instance.reward_max * reward_w)
            self.reward_min += dt_float(reward_instance.reward_min * reward_w)
        env.reward_range = self.get_range()

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        res = dt_float(0.0)
        # Loop over registered rewards
        for key, reward in self.rewards.items():
            r_instance = reward["instance"]
            # Call individual reward
            r = r_instance(action, env, has_error, is_done, is_illegal, is_ambiguous)
            # Sum by weighted result
            w = dt_float(reward["weight"])
            res += dt_float(r) * w
        # Return total sum
        return res

    def close(self):
        for key, reward in self.rewards.items():
            reward["instance"].close()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Agent.baseAgent import BaseAgent


class DeltaRedispatchRandomAgent(BaseAgent):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Used for test. Prefer using a random agent by selecting only the redispatching action
        that you want.

    This agent will perform some redispatch of a given amount among randomly selected dispatchable
    generators.

    Parameters
    ----------
    action_space: :class:`grid2op.Action.ActionSpace`
         the Grid2Op action space

    n_gens_to_redispatch: `int`
      The maximum number of dispatchable generators to play with

    redispatching_delta: `float`
      The redispatching MW value used in both directions

    """

    def __init__(self, action_space, n_gens_to_redispatch=2, redispatching_delta=1.0):
        super().__init__(action_space)
        self.desired_actions = []

        # Get all generators IDs
        gens_ids = np.arange(self.action_space.n_gen, dtype=int)
        # Filter out non resipatchable IDs
        gens_redisp = gens_ids[self.action_space.gen_redispatchable]
        # Cut if needed
        if len(gens_redisp) > n_gens_to_redispatch:
            gens_redisp = gens_redisp[0:n_gens_to_redispatch]

        # Register do_nothing action
        self.desired_actions.append(self.action_space({}))

        # Register 2 actions per generator
        # (increase or decrease by the delta)
        for gen_id in gens_redisp:
            # Create action redispatch by opposite delta
            act1 = self.action_space(
                {"redispatch": [(gen_id, -float(redispatching_delta))]}
            )

            # Create action redispatch by delta
            act2 = self.action_space(
                {"redispatch": [(gen_id, float(redispatching_delta))]}
            )

            # Register this generator actions
            self.desired_actions.append(act1)
            self.desired_actions.append(act2)

    def act(self, observation, reward, done=False):
        act = self.space_prng.choice(self.desired_actions)
        return act



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.dtypes import dt_int
from grid2op.Chronics import Multifolder
from grid2op.gym_compat.utils import GYM_AVAILABLE, GYMNASIUM_AVAILABLE
from grid2op.gym_compat.utils import (check_gym_version, sample_seed)
    
    
def conditional_decorator(condition):
    def decorator(func):
        if condition:
            # Return the function unchanged, not decorated.
            return func
        return NotImplementedError()  # anything that is not a callbe anyway
    return decorator


class __AuxGymEnv:
    """
    fully implements the openAI gym API by using the :class:`GymActionSpace` and :class:`GymObservationSpace`
    for compliance with openAI gym.

    They can handle action_space_converter or observation_space converter to change the representation of data
    that will be fed to the agent.  #TODO

    .. warning::
        The `gym` package has some breaking API change since its version 0.26. Depending on the version installed,
        we attempted, in grid2op, to maintain compatibility both with former version and later one. This makes this
        class behave differently depending on the version of gym you have installed !
        
        The main changes involve the functions `env.step` and `env.reset`
        
    If you want to use the same version of the GymEnv regardless of the gym / gymnasium version installed you can use:
    
    - :class:`GymnasiumEnv` if `gymnasium` is available
    - :class:`GymEnv_Legacy` for gym < 0.26
    - :class:`GymEnv_Modern` for gym >= 0.26

    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`GymEnv` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`GymnasiumEnv`), otherwise it will
          inherit from gym (and will be exactly :class:`GymEnv_Legacy` - gym < 0.26- 
          or :class:`GymEnv_Modern` - for gym >= 0.26)
        - :class:`GymnasiumEnv` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`GymEnv_Legacy` and :class:`GymEnv_Modern` will inherit from gym if it's 
          available and never from from gymnasium
        
        See :ref:`gymnasium_gym` for more information
        
    Notes
    ------
    The environment passed as input is copied. It is not modified by this "gym environment"

    Examples
    --------
    This can be used like:

    .. code-block:: python

        import grid2op
        from grid2op.gym_compat import GymEnv

        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)
        gym_env = GymEnv(env)  # is a gym environment properly inheriting from gym.Env !

    There are a few difference between "raw" grid2op environment and gymnasium environments.
    
    One of the major difference is that, to our knowledge, gymnasium does not support the 
    `simulate` feature (which allows an agent to test the impact of a given action 
    on the grid without having to perform a `step` see :ref:`model_based_rl` for more information)
    [NB if you know or better are developping some "model based RL library" let us know !]
    
    Another difference is in the way to do some actions. In grid2op, actions are a dedicated class
    and can be made with an `action_space` and a dictionary, or  using the properties of the action
    class.
    
    In gym, there are no specific representations of the action class. More precisely, for each action
    type (:class:`MultiDiscreteActSpace`, :class:`DiscreteActSpace`, :class:`BoxGymActSpace` or 
    :class:`GymActionSpace`) there is a way to encode it. For example, by default (:class:`GymActionSpace`)
    an action is represented through an OrderedDict (`from collection import OrderedDict`)
    """

    def __init__(self, env_init, shuffle_chronics=True, render_mode="rgb_array"):
        check_gym_version(type(self)._gymnasium)
        self.init_env = env_init.copy()
        self.action_space = type(self)._ActionSpaceType(self.init_env)
        self.observation_space = type(self)._ObservationSpaceType(self.init_env)
        self.reward_range = self.init_env.reward_range
        self.metadata = self.init_env.metadata
        self.init_env.render_mode = render_mode
        self._shuffle_chronics = shuffle_chronics
            
        super().__init__()  # super should reference either gym.Env or gymnasium.Env
        if not hasattr(self, "_np_random"):
            # for older version of gym it does not exist
            self._np_random = np.random.RandomState()
        
    def _aux_step(self, gym_action):
        # used for gym < 0.26
        g2op_act = self.action_space.from_gym(gym_action)
        g2op_obs, reward, done, info = self.init_env.step(g2op_act)
        gym_obs = self.observation_space.to_gym(g2op_obs)
        return gym_obs, float(reward), done, info
    
    def _aux_step_new(self, gym_action):
        # used for gym >= 0.26
        # TODO refacto with _aux_step
        g2op_act = self.action_space.from_gym(gym_action)
        g2op_obs, reward, terminated, info = self.init_env.step(g2op_act)
        gym_obs = self.observation_space.to_gym(g2op_obs)
        truncated = False # see https://github.com/openai/gym/pull/2752
        return gym_obs, float(reward), terminated, truncated, info

    def _aux_reset(self, seed=None, return_info=None, options=None):
        # used for gym < 0.26
        if self._shuffle_chronics and isinstance(
            self.init_env.chronics_handler.real_data, Multifolder
        ):
            self.init_env.chronics_handler.sample_next_chronics()
         
        if seed is not None:
            seed_, next_seed, underlying_env_seeds = self._aux_seed(seed)
            
        g2op_obs = self.init_env.reset()
        gym_obs = self.observation_space.to_gym(g2op_obs)
            
        if return_info:
            chron_id = self.init_env.chronics_handler.get_id()
            info = {"time serie id": chron_id}
            if seed is not None:
                info["seed"] = seed
                info["grid2op_env_seed"] = next_seed
                info["underlying_env_seeds"] = underlying_env_seeds
            return gym_obs, info
        else:
            return gym_obs

    def _aux_reset_new(self, seed=None, options=None):
        # used for gym > 0.26
        if self._shuffle_chronics and isinstance(
            self.init_env.chronics_handler.real_data, Multifolder
        ):
            self.init_env.chronics_handler.sample_next_chronics()
        
        super().reset(seed=seed)    
        if seed is not None:
            self._aux_seed_spaces()
            seed, next_seed, underlying_env_seeds = self._aux_seed_g2op(seed)
            
        g2op_obs = self.init_env.reset()
        gym_obs = self.observation_space.to_gym(g2op_obs)
            
        chron_id = self.init_env.chronics_handler.get_id()
        info = {"time serie id": chron_id}
        if seed is not None:
            info["seed"] = seed
            info["grid2op_env_seed"] = next_seed
            info["underlying_env_seeds"] = underlying_env_seeds
        return gym_obs, info
        
    def render(self):
        """for compatibility with open ai gym render function"""
        return self.init_env.render()

    def close(self):
        if hasattr(self, "init_env") and self.init_env is not None:
            self.init_env.close()
            del self.init_env
        self.init_env = None
        
        if hasattr(self, "action_space") and self.action_space is not None:
            self.action_space.close()
        self.action_space = None
        
        if hasattr(self, "observation_space") and self.observation_space is not None:
            self.observation_space.close()
        self.observation_space = None

    def _aux_seed_spaces(self):
        max_ = np.iinfo(dt_int).max 
        next_seed = sample_seed(max_, self._np_random)
        self.action_space.seed(next_seed)
        next_seed = sample_seed(max_, self._np_random)
        self.observation_space.seed(next_seed)
            
    def _aux_seed_g2op(self, seed):
            # then seed the underlying grid2op env
            max_ = np.iinfo(dt_int).max 
            next_seed = sample_seed(max_, self._np_random)
            underlying_env_seeds = self.init_env.seed(next_seed)
            return seed, next_seed, underlying_env_seeds
        
    def _aux_seed(self, seed=None):
        # deprecated in gym >=0.26
        if seed is not None:
            # seed the gym env
            super().seed(seed)
            self._np_random.seed(seed)
            self._aux_seed_spaces()
            return self._aux_seed_g2op(seed)
        return None, None, None

    def __del__(self):
        # delete possible dangling reference
        self.close()

if GYM_AVAILABLE:
    from gym import Env as LegacyGymEnv
    from grid2op.gym_compat.gym_obs_space import LegacyGymObservationSpace
    from grid2op.gym_compat.gym_act_space import LegacyGymActionSpace
    _AuxGymEnv = type("_AuxGymEnv",
                      (__AuxGymEnv, LegacyGymEnv),
                      {"_gymnasium": False,
                       "_ActionSpaceType": LegacyGymActionSpace,
                       "_ObservationSpaceType": LegacyGymObservationSpace,
                       "__module__": __name__})
    _AuxGymEnv.__doc__ = __AuxGymEnv.__doc__
    class GymEnv_Legacy(_AuxGymEnv):
        # for old version of gym        
        def reset(self, *args, **kwargs):
            return self._aux_reset(*args, **kwargs)

        def step(self, action):
            return self._aux_step(action)

        def seed(self, seed):
            # defined only on some cases
            return self._aux_seed(seed)


    class GymEnv_Modern(_AuxGymEnv):
        # for new version of gym
        def reset(self,
                *,
                seed=None,
                options=None,):
            return self._aux_reset_new(seed, options)

        def step(self, action):
            return self._aux_step_new(action)
    GymEnv_Legacy.__doc__ = __AuxGymEnv.__doc__
    GymEnv_Modern.__doc__ = __AuxGymEnv.__doc__


if GYMNASIUM_AVAILABLE:
    from gymnasium import Env
    from grid2op.gym_compat.gym_act_space import GymnasiumActionSpace
    from grid2op.gym_compat.gym_obs_space import GymnasiumObservationSpace
    _AuxGymnasiumEnv = type("_AuxGymnasiumEnv",
                            (__AuxGymEnv, Env),
                            {"_gymnasium": True,
                             "_ActionSpaceType": GymnasiumActionSpace,
                             "_ObservationSpaceType": GymnasiumObservationSpace,
                             "__module__": __name__})
    _AuxGymnasiumEnv.__doc__ = __AuxGymEnv.__doc__
    
    class GymnasiumEnv(_AuxGymnasiumEnv):
        # for new version of gym
        def reset(self,
                *,
                seed=None,
                options=None,):
            return self._aux_reset_new(seed, options)

        def step(self, action):
            return self._aux_step_new(action)
    GymnasiumEnv.__doc__ = __AuxGymEnv.__doc__


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from datetime import datetime, timedelta

from grid2op.dtypes import dt_int
from grid2op.Chronics.gridValue import GridValue


class ChangeNothing(GridValue):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Do not attempt to create an object of this class. This is initialized by the environment
        at its creation.

    This set of class is mainly internal.

    We don't recommend you, unless you want to code a custom "chroncis class" to change anything
    on these classes.

    This class is the most basic class to modify a powergrid values.
    It does nothing aside from increasing :attr:`GridValue.max_iter` and the :attr:`GridValue.current_datetime`.

    Examples
    --------

    Usage example, for what you don't really have to do:

    .. code-block:: python

        import grid2op
        from grid2op.Chronics import ChangeNothing

        env_name = "l2rpn_case14_sandbox"  # or any other name
        # env = grid2op.make(env_name, data_feeding_kwargs={"gridvalueClass": ChangeNothing})
        env = grid2op.make(env_name, chronics_class=ChangeNothing)

    It can also be used with the "blank" environment:

    .. code-block:: python

        import grid2op
        from grid2op.Chronics import ChangeNothing
        env = grid2op.make("blank",
                           test=True,
                           grid_path=EXAMPLE_CASEFILE,
                           chronics_class=ChangeNothing,
                           action_class=TopologyAndDispatchAction)
        
    """
    MULTI_CHRONICS = False
    def __init__(
        self,
        time_interval=timedelta(minutes=5),
        max_iter=-1,
        start_datetime=datetime(year=2019, month=1, day=1),
        chunk_size=None,
        **kwargs
    ):
        GridValue.__init__(
            self,
            time_interval=time_interval,
            max_iter=max_iter,
            start_datetime=start_datetime,
            chunk_size=chunk_size,
        )
        self.n_gen = None
        self.n_load = None
        self.n_line = None
        
        self.maintenance_time = None
        self.maintenance_duration = None
        self.hazard_duration = None
        
    def initialize(
        self,
        order_backend_loads,
        order_backend_prods,
        order_backend_lines,
        order_backend_subs,
        names_chronics_to_backend=None,
    ):
        self.n_gen = len(order_backend_prods)
        self.n_load = len(order_backend_loads)
        self.n_line = len(order_backend_lines)
        self.curr_iter = 0

        self.maintenance_time = np.zeros(shape=(self.n_line,), dtype=dt_int) - 1
        self.maintenance_duration = np.zeros(shape=(self.n_line,), dtype=dt_int)
        self.hazard_duration = np.zeros(shape=(self.n_line,), dtype=dt_int)

    def load_next(self):
        self.current_datetime += self.time_interval
        self.curr_iter += 1
        return (
            self.current_datetime,
            {},
            self.maintenance_time,
            self.maintenance_duration,
            self.hazard_duration,
            None,
        )

    def check_validity(self, backend):
        return True

    def next_chronics(self):
        self.current_datetime = self.start_datetime
        self.curr_iter = 0



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import copy
import os

import requests
import time
import grid2op
import numpy as np
from tqdm import tqdm
import asyncio
import aiohttp
import warnings
import subprocess
import sys

try:
    import ujson

    requests.models.complexjson = ujson
except ImportError as exc_:
    warnings.warn(
        "usjon is not installed. You could potentially get huge benefit if installing it"
    )

ERROR_NO_200 = "error due to not receiving 200 status"

NB_SUB_ENV = 4
ENV_NAME = "l2rpn_neurips_2020_track2_small"
SYNCH = True
NB_step = 300


PORTS = [3000 + i for i in range(NB_SUB_ENV)]  # TODO start them on the fly


class MultiEnvServer:
    def __init__(self, ports=PORTS, env_name=ENV_NAME, address="http://127.0.0.1"):
        warnings.warn(
            "This is an alpha feature and has absolutely not interest at the moment. Do not use unless "
            "you want to improve this feature yourself (-:"
        )
        self.my_procs = []
        for port in ports:
            p_ = subprocess.Popen(
                [
                    sys.executable,
                    "/home/benjamin/Documents/grid2op_dev/grid2op/rest_server/app.py",
                    "--port",
                    f"{port}",
                ],
                env=os.environ,
                # stdout=subprocess.DEVNULL,  # TODO logger
                # stderr=subprocess.DEVNULL  # TODO logger
            )
            self.my_procs.append(p_)

        self.nb_env = len(ports)
        self.ports = ports
        self.address = address
        self.li_urls = ["{}:{}".format(address, port) for port in ports]
        self.env_name = env_name
        self._local_env = grid2op.make(env_name)
        if SYNCH:
            self.session = requests.session()
        else:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)

        self.action_space = self._local_env.action_space
        self.observation_space = self._local_env.observation_space

        if SYNCH:
            answ_json = self._make_env_synch()
        else:
            answ_json = self.loop.run_until_complete(self._make_env_asynch())

        self.env_id = [int(el["id"]) for el in answ_json]
        self.obs = [el["obs"] for el in answ_json]

    def _make_env_synch(self):
        answs = []
        for url in self.li_urls:
            resp = self.session.get(f"{url}/make/{self.env_name}")
            answs.append(resp)
        import pdb

        pdb.set_trace()
        assert np.all(np.array([el.status_code for el in answs]) == 200), ERROR_NO_200
        answ_json = [el.json() for el in answs]
        return answ_json

    async def _make_env_asynch(self):
        answ_json = []
        async with aiohttp.ClientSession() as session:
            for url in self.li_urls:
                async with session.get(f"{url}/make/{self.env_name}") as resp:
                    if resp.status != 200:
                        raise RuntimeError(ERROR_NO_200)
                    answ_json.append(await resp.json())
        return answ_json

    def _step_synch(self, acts):
        answs = []
        for url, id_env, act in zip(self.li_urls, self.env_id, acts):
            resp = self.session.post(
                f"{url}/step/{self.env_name}/{id_env}", json={"action": act.to_json()}
            )
            answs.append(resp)
        answs = [el.json() for el in answs]
        return answs

    async def _step_asynch(self, acts):
        answs = []
        async with aiohttp.ClientSession() as session:
            for url, id_env, act in zip(self.li_urls, self.env_id, acts):
                async with session.post(
                    f"{url}/step/{self.env_name}/{id_env}",
                    json={"action": act.to_json()},
                ) as resp:
                    if resp.status != 200:
                        raise RuntimeError(ERROR_NO_200)
                    answs.append(await resp.json())
        return answs

    def step(self, acts):
        if SYNCH:
            answ_json = self._step_synch(acts)
        else:
            answ_json = self.loop.run_until_complete(self._step_asynch(acts))

        obss = [el["obs"] for el in answ_json]
        rewards = [el["reward"] for el in answ_json]
        info = [el["info"] for el in answ_json]
        done = [el["done"] for el in answ_json]
        return obss, rewards, done, info

    def close(self):
        """close all the opened port"""
        for p_ in self.my_procs:
            p_.terminate()
            p_.kill()


if __name__ == "__main__":
    multi_env = MultiEnvServer()
    try:
        beg = time.perf_counter()
        for _ in tqdm(range(NB_step)):
            obs, reward, done, info = multi_env.step(
                [multi_env.action_space() for _ in range(multi_env.nb_env)]
            )
        end = time.perf_counter()
    finally:
        multi_env.close()
    print(
        f"Using {'synchronous' if SYNCH else 'asyncio'}, it took {end-beg:.2f}s to make {NB_step} steps "
        f"on {ENV_NAME} using {len(PORTS)} sub environment(s)."
    )



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import importlib.util
import numpy as np
import json
import warnings

from grid2op.Environment import Environment
from grid2op.Backend import Backend, PandaPowerBackend
from grid2op.Opponent.OpponentSpace import OpponentSpace
from grid2op.Parameters import Parameters
from grid2op.Chronics import ChronicsHandler, ChangeNothing, FromNPY, FromChronix2grid
from grid2op.Chronics import GridStateFromFile, GridValue
from grid2op.Action import BaseAction, DontAct
from grid2op.Exceptions import *
from grid2op.Observation import CompleteObservation, BaseObservation
from grid2op.Reward import BaseReward, L2RPNReward
from grid2op.Rules import BaseRules, DefaultRules
from grid2op.VoltageControler import ControlVoltageFromFile
from grid2op.Opponent import BaseOpponent, BaseActionBudget, NeverAttackBudget
from grid2op.operator_attention import LinearAttentionBudget

from grid2op.MakeEnv.get_default_aux import _get_default_aux

DIFFICULTY_NAME = "difficulty"
CHALLENGE_NAME = "competition"
ERR_MSG_KWARGS = {
    "backend": 'The backend of the environment (keyword "backend") must be an instance of grid2op.Backend',
    "observation_class": 'The type of observation of the environment (keyword "observation_class")'
    " must be a subclass of grid2op.BaseObservation",
    "param": 'The parameters of the environment (keyword "param") must be an instance of grid2op.Parameters',
    "gamerules_class": 'The type of rules of the environment (keyword "gamerules_class")'
    " must be a subclass of grid2op.BaseRules",
    "reward_class": 'The type of reward in the environment (keyword "reward_class") must be a subclass of '
    "grid2op.BaseReward",
    "action_class": 'The type of action of the environment (keyword "action_class") must be a subclass of '
    "grid2op.BaseAction",
    "data_feeding_kwargs": "The argument to build the data generation process [chronics]"
    '  (keyword "data_feeding_kwargs") should be a dictionnary.',
    "chronics_class": 'The argument to build the data generation process [chronics] (keyword "chronics_class")'
    " should be a class that inherit grid2op.Chronics.GridValue.",
    "chronics_handler": 'The argument to build the data generation process [chronics] (keyword "data_feeding")'
    " should be a class that inherit grid2op.ChronicsHandler.ChronicsHandler.",
    "voltagecontroler_class": "The argument to build the online controler for chronics (keyword "
    '"volagecontroler_class")'
    " should be a class that inherit grid2op.VoltageControler.ControlVoltageFromFile.",
    "names_chronics_to_grid": 'The converter between names (keyword "names_chronics_to_backend") '
    "should be a dictionnary.",
    "other_rewards": 'The argument to build the online controler for chronics (keyword "other_rewards") '
    "should be dictionary.",
    "chronics_path": 'The path where the data is located (keyword "chronics_path") should be a string.',
    "grid_path": 'The path where the grid is located (keyword "grid_path") should be a string.',
    "opponent_space_type": 'The argument used to build the opponent space (expects a type / class and not an instance of that type)',
    "opponent_action_class": 'The argument used to build the "opponent_action_class" should be a class that '
    'inherit from "BaseAction"',
    "opponent_class": 'The argument used to build the "opponent_class" should be a class that '
    'inherit from "BaseOpponent"',
    "opponent_attack_duration": "The number of time steps an attack from the opponent lasts",
    "opponent_attack_cooldown": "The number of time steps the opponent as to wait for an attack",
    "opponent_init_budget": 'The initial budget of the opponent "opponent_init_budget" should be a float',
    "opponent_budget_class": 'The opponent budget class ("opponent_budget_class") should derive from '
    '"BaseActionBudget".',
    "opponent_budget_per_ts": 'The increase of the opponent\'s budget ("opponent_budget_per_ts") should be a float.',
    "kwargs_opponent": "The extra kwargs argument used to properly initialized the opponent "
    '("kwargs_opponent") should '
    "be a dictionary.",
    "has_attention_budget": 'The "has_attention_budget" key word argument should be a flag indicating whether '
    "you want this feature or not. It should be a boolean.",
    "attention_budget_class": 'The attention budget class ("attention_budget_class") should derive from '
    '"LinearAttentionBudget".',
    "kwargs_attention_budget": "The extra kwargs argument used to properly initialized the attention budget "
    '("kwargs_attention_budget") should '
    "be a dictionary.",
    DIFFICULTY_NAME: "Unknown difficulty level {difficulty} for this environment. Authorized difficulties are "
    "{difficulties}",
    "kwargs_observation": "The extra kwargs argument used to properly initialized each observations "
    '("kwargs_observation") should '
    "be a dictionary.",
    "observation_backend_class": ("The class used to build the observation backend (used for Simulator "
                                  "obs.simulate and obs.get_forecasted_env). If provided, this should "
                                  "be a type / class and not an instance of this class. (by default it's None)"),
    "observation_backend_kwargs": ("key-word arguments to build the observation backend (used for Simulator, "
    " obs.simulate and obs.get_forecasted_env). This should be a dictionnary. (by default it's None)")
}

NAME_CHRONICS_FOLDER = "chronics"
NAME_GRID_FILE = "grid.json"
NAME_GRID_LAYOUT_FILE = "grid_layout.json"
NAME_CONFIG_FILE = "config.py"


def _check_kwargs(kwargs):
    for el in kwargs:
        if el not in ERR_MSG_KWARGS.keys():
            raise EnvError(
                'Unknown keyword argument "{}" used to create an Environment. '
                "No Environment will be created. "
                "Accepted keyword arguments are {}".format(el, ERR_MSG_KWARGS.keys())
            )


def _check_path(path, info):
    if path is None or os.path.exists(path) is False:
        raise EnvError("Cannot find {}. {}".format(path, info))


def make_from_dataset_path(
    dataset_path="/",
    logger=None,
    experimental_read_from_local_dir=False,
    _add_to_name="",
    _compat_glop_version=None,
    **kwargs,
) -> Environment:
    """
    INTERNAL USE ONLY

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Prefer using the :func:`grid2op.make` function.


    This function is a shortcut to rapidly create environments within the grid2op Framework. We don't
    recommend using directly this function. Prefer using the :func:`make` function.

    It mimic the ``gym.make`` function.

    .. _Parameters-make-from-path:

    Parameters
    ----------

    dataset_path: ``str``
        Path to the dataset folder

    logger:
        Something to pass to grid2op environment to be used as logger.

    param: ``grid2op.Parameters.Parameters``, optional
        Type of parameters used for the Environment. Parameters defines how the powergrid problem is cast into an
        markov decision process, and some internal

    backend: ``grid2op.Backend.Backend``, optional
        The backend to use for the computation. If provided, it must be an instance of :class:`grid2op.Backend.Backend`.

    action_class: ``type``, optional
        Type of BaseAction the BaseAgent will be able to perform.
        If provided, it must be a subclass of :class:`grid2op.BaseAction.BaseAction`

    observation_class: ``type``, optional
        Type of BaseObservation the BaseAgent will receive.
        If provided, It must be a subclass of :class:`grid2op.BaseAction.BaseObservation`

    reward_class: ``type``, optional
        Type of reward signal the BaseAgent will receive.
        If provided, It must be a subclass of :class:`grid2op.BaseReward.BaseReward`

    other_rewards: ``dict``, optional
        Used to additional information than the "info" returned value after a call to env.step.

    gamerules_class: ``type``, optional
        Type of "Rules" the BaseAgent need to comply with. Rules are here to model some operational constraints.
        If provided, It must be a subclass of :class:`grid2op.RulesChecker.BaseRules`

    data_feeding_kwargs: ``dict``, optional
        Dictionnary that is used to build the `data_feeding` (chronics) objects.

    chronics_class: ``type``, optional
        The type of chronics that represents the dynamics of the Environment created. Usually they come from different
        folders.

    data_feeding: ``type``, optional
        The type of chronics handler you want to use.

    volagecontroler_class: ``type``, optional
        The type of :class:`grid2op.VoltageControler.VoltageControler` to use, it defaults to

    chronics_path: ``str``
        Path where to look for the chronics dataset (optional)

    grid_path: ``str``, optional
        The path where the powergrid is located.
        If provided it must be a string, and point to a valid file present on the hard drive.

    difficulty: ``str``, optional
        the difficulty level. If present it starts from "0" the "easiest" but least realistic mode. In the case of the
        dataset being used in the l2rpn competition, the level used for the competition is "competition" ("hardest" and
        most realistic mode). If multiple difficulty levels are available, the most realistic one
        (the "hardest") is the default choice.

    opponent_space_type: ``type``, optional
        The type of opponent space to use. If provided, it must be a subclass of `OpponentSpace`.
        
    opponent_action_class: ``type``, optional
        The action class used for the opponent. The opponent will not be able to use action that are invalid with
        the given action class provided. It defaults to :class:`grid2op.Action.DontAct` which forbid any type
        of action possible.

    opponent_class: ``type``, optional
        The opponent class to use. The default class is :class:`grid2op.Opponent.BaseOpponent` which is a type
        of opponents that does nothing.

    opponent_init_budget: ``float``, optional
        The initial budget of the opponent. It defaults to 0.0 which means the opponent cannot perform any action
        if this is not modified.

    opponent_attack_duration: ``int``, optional
        The number of time steps an attack from the opponent lasts.

    opponent_attack_cooldown: ``int``, optional
        The number of time steps the opponent as to wait for an attack.

    opponent_budget_per_ts: ``float``, optional
        The increase of the opponent budget per time step. Each time step the opponent see its budget increase. It
        defaults to 0.0.

    opponent_budget_class: ``type``, optional
        defaults: :class:`grid2op.Opponent.UnlimitedBudget`

    kwargs_observation: ``dict``
        Key words used to initialize the observation. For example, in case of NoisyObservation, 
        it might be the standar error for each underlying distribution. It might
        be more complicated for other type of custom observations but should be
        deep copiable.

        Each observation will be initialized (by the observation_space) with:

        .. code-block:: python
        
            obs = observation_class(obs_env=self.obs_env,
                                    action_helper=self.action_helper_env,
                                    random_prng=self.space_prng,
                                    **kwargs_observation  # <- this kwargs is used here
                                   )

    observation_backend_class:
        The class used to build the observation backend (used for Simulator 
        obs.simulate and obs.get_forecasted_env). If provided, this should 
        be a type / class and not an instance of this class. (by default it's None)
        
    observation_backend_kwargs:
        The key-word arguments to build the observation backend (used for Simulator, 
        obs.simulate and obs.get_forecasted_env). This should be a dictionnary. 
        (by default it's None)
    
    _add_to_name:
        Internal, used for test only. Do not attempt to modify under any circumstances.

    _compat_glop_version:
        Internal, used for test only. Do not attempt to modify under any circumstances.

    # TODO update doc with attention budget

    Returns
    -------
    env: :class:`grid2op.Environment.Environment`
        The created environment with the given properties.

    """    
    # Compute and find root folder
    _check_path(dataset_path, "Dataset root directory")
    dataset_path_abs = os.path.abspath(dataset_path)

    # Compute env name from directory name
    name_env = os.path.split(dataset_path_abs)[1]

    # Compute and find chronics folder
    chronics_path = _get_default_aux(
        "chronics_path",
        kwargs,
        defaultClassApp=str,
        defaultinstance="",
        msg_error=ERR_MSG_KWARGS["chronics_path"],
    )
    if chronics_path == "":
        # if no "chronics_path" argument is provided, look into the "chronics" folder
        chronics_path_abs = os.path.abspath(
            os.path.join(dataset_path_abs, NAME_CHRONICS_FOLDER)
        )
    else:
        # otherwise use it
        chronics_path_abs = os.path.abspath(chronics_path)
    exc_chronics = None
    try:
        _check_path(chronics_path_abs, "Dataset chronics folder")
    except Exception as exc_:
        exc_chronics = exc_

    # Compute and find backend/grid file
    grid_path = _get_default_aux(
        "grid_path",
        kwargs,
        defaultClassApp=str,
        defaultinstance="",
        msg_error=ERR_MSG_KWARGS["grid_path"],
    )
    if grid_path == "":
        grid_path_abs = os.path.abspath(os.path.join(dataset_path_abs, NAME_GRID_FILE))
    else:
        grid_path_abs = os.path.abspath(grid_path)
    _check_path(grid_path_abs, "Dataset power flow solver configuration")

    # Compute and find grid layout file
    grid_layout_path_abs = os.path.abspath(
        os.path.join(dataset_path_abs, NAME_GRID_LAYOUT_FILE)
    )
    try:
        _check_path(grid_layout_path_abs, "Dataset grid layout")
    except EnvError as exc_:
        warnings.warn(
            f'Impossible to load the coordinate of the substation with error: "{exc_}". Expect some issue '
            f"if you attempt to plot the grid."
        )

    # Check provided config overrides are valid
    _check_kwargs(kwargs)

    # Compute and find config file
    config_path_abs = os.path.abspath(os.path.join(dataset_path_abs, NAME_CONFIG_FILE))
    _check_path(config_path_abs, "Dataset environment configuration")

    # Read config file
    try:
        spec = importlib.util.spec_from_file_location("config.config", config_path_abs)
        config_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(config_module)
        config_data = config_module.config
    except Exception as exc_:
        print(exc_)
        raise EnvError(
            "Invalid dataset config file: {}".format(config_path_abs)
        ) from None

    # Get graph layout
    graph_layout = None
    try:
        with open(grid_layout_path_abs) as layout_fp:
            graph_layout = json.load(layout_fp)
    except Exception as exc_:
        warnings.warn(
            "Dataset {} doesn't have a valid graph layout. Expect some failures when attempting "
            "to plot the grid. Error was: {}".format(config_path_abs, exc_)
        )

    # Get thermal limits
    thermal_limits = None
    if "thermal_limits" in config_data:
        thermal_limits = config_data["thermal_limits"]

    # Get chronics_to_backend
    name_converter = None
    if "names_chronics_to_grid" in config_data:
        name_converter = config_data["names_chronics_to_grid"]
    if name_converter is None:
        name_converter = {}
        is_none = True
    else:
        is_none = False
    names_chronics_to_backend = _get_default_aux(
        "names_chronics_to_backend",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=name_converter,
        msg_error=ERR_MSG_KWARGS["names_chronics_to_grid"],
    )
    if is_none and names_chronics_to_backend  == {}:
        names_chronics_to_backend = None
        
    # Get default backend class
    backend_class_cfg = PandaPowerBackend
    if "backend_class" in config_data and config_data["backend_class"] is not None:
        backend_class_cfg = config_data["backend_class"]
    ## Create the backend, to compute the powerflow
    backend = _get_default_aux(
        "backend",
        kwargs,
        defaultClass=backend_class_cfg,
        defaultClassApp=Backend,
        msg_error=ERR_MSG_KWARGS["backend"],
    )

    # Get default observation class
    observation_class_cfg = CompleteObservation
    if (
        "observation_class" in config_data
        and config_data["observation_class"] is not None
    ):
        observation_class_cfg = config_data["observation_class"]
    ## Setup the type of observation the agent will receive
    observation_class = _get_default_aux(
        "observation_class",
        kwargs,
        defaultClass=observation_class_cfg,
        isclass=True,
        defaultClassApp=BaseObservation,
        msg_error=ERR_MSG_KWARGS["observation_class"],
    )

    ## Create the parameters of the game, thermal limits threshold,
    # simulate cascading failure, powerflow mode etc. (the gamification of the game)
    if "param" in kwargs:
        param = _get_default_aux(
            "param",
            kwargs,
            defaultClass=Parameters,
            defaultClassApp=Parameters,
            msg_error=ERR_MSG_KWARGS["param"],
        )
    else:
        # param is not in kwargs
        param = Parameters()
        json_path = os.path.join(dataset_path_abs, "difficulty_levels.json")
        if os.path.exists(json_path):
            with open(json_path, "r", encoding="utf-8") as f:
                dict_ = json.load(f)
            available_parameters = sorted(dict_.keys())
            if DIFFICULTY_NAME in kwargs:
                # player enters a difficulty levels
                my_difficulty = kwargs[DIFFICULTY_NAME]
                try:
                    my_difficulty = str(my_difficulty)
                except Exception as exc_:
                    raise EnvError(
                        "Impossible to convert your difficulty into a valid string. Please make sure to "
                        'pass a string (eg "2") and not something else (eg. int(2)) as a difficulty.'
                        "Error was \n{}".format(exc_)
                    )
                if my_difficulty in dict_:
                    param.init_from_dict(dict_[my_difficulty])
                else:
                    raise EnvError(
                        ERR_MSG_KWARGS[DIFFICULTY_NAME].format(
                            difficulty=my_difficulty, difficulties=available_parameters
                        )
                    )
            else:
                # no difficulty name provided, i need to chose the most suited one
                if CHALLENGE_NAME in dict_:
                    param.init_from_dict(dict_[CHALLENGE_NAME])
                else:
                    # i chose the most difficult one
                    available_parameters_int = {}
                    for el in available_parameters:
                        try:
                            int_ = int(el)
                            available_parameters_int[int_] = el
                        except:
                            pass
                    max_ = np.max(list(available_parameters_int.keys()))
                    keys_ = available_parameters_int[max_]
                    param.init_from_dict(dict_[keys_])
        else:
            json_path = os.path.join(dataset_path_abs, "parameters.json")
            if os.path.exists(json_path):
                param.init_from_json(json_path)

    # Get default rules class
    rules_class_cfg = DefaultRules
    if "rules_class" in config_data and config_data["rules_class"] is not None:
        warnings.warn("You used the deprecated rules_class in your config. Please change its "
                      "name to 'gamerules_class' to mimic the grid2op.make kwargs.")
        rules_class_cfg = config_data["rules_class"]
    if "gamerules_class" in config_data and config_data["gamerules_class"] is not None:
        rules_class_cfg = config_data["gamerules_class"]
        
    ## Create the rules of the game (mimic the operationnal constraints)
    gamerules_class = _get_default_aux(
        "gamerules_class",
        kwargs,
        defaultClass=rules_class_cfg,
        defaultClassApp=BaseRules,
        msg_error=ERR_MSG_KWARGS["gamerules_class"],
        isclass=None,
    )

    # Get default reward class
    reward_class_cfg = L2RPNReward
    if "reward_class" in config_data and config_data["reward_class"] is not None:
        reward_class_cfg = config_data["reward_class"]

    ## Setup the reward the agent will receive
    reward_class = _get_default_aux(
        "reward_class",
        kwargs,
        defaultClass=reward_class_cfg,
        defaultClassApp=BaseReward,
        msg_error=ERR_MSG_KWARGS["reward_class"],
        isclass=None,
    )

    # Get default BaseAction class
    action_class_cfg = BaseAction
    if "action_class" in config_data and config_data["action_class"] is not None:
        action_class_cfg = config_data["action_class"]
    ## Setup the type of action the BaseAgent can perform
    action_class = _get_default_aux(
        "action_class",
        kwargs,
        defaultClass=action_class_cfg,
        defaultClassApp=BaseAction,
        msg_error=ERR_MSG_KWARGS["action_class"],
        isclass=True,
    )

    # Get default Voltage class
    voltage_class_cfg = ControlVoltageFromFile
    if "voltage_class" in config_data and config_data["voltage_class"] is not None:
        voltage_class_cfg = config_data["voltage_class"]
    ### Create controler for voltages
    volagecontroler_class = _get_default_aux(
        "volagecontroler_class",
        kwargs,
        defaultClassApp=voltage_class_cfg,
        defaultClass=ControlVoltageFromFile,
        msg_error=ERR_MSG_KWARGS["voltagecontroler_class"],
        isclass=True,
    )

    # Get default Chronics class
    chronics_class_cfg = ChangeNothing
    if "chronics_class" in config_data and config_data["chronics_class"] is not None:
        chronics_class_cfg = config_data["chronics_class"]
        
    # Get default Grid class
    grid_value_class_cfg = GridStateFromFile
    if (
        "grid_value_class" in config_data
        and config_data["grid_value_class"] is not None
    ):
        grid_value_class_cfg = config_data["grid_value_class"]

    ## the chronics to use
    ### the arguments used to build the data, note that the arguments must be compatible with the chronics class
    default_chronics_kwargs = {
        "path": chronics_path_abs,
        "chronicsClass": chronics_class_cfg,
        # "gridvalueClass": grid_value_class_cfg,
    }

    if "data_feeding_kwargs" in config_data and config_data["data_feeding_kwargs"] is not None:
        dfkwargs_cfg = config_data["data_feeding_kwargs"]
        for el in dfkwargs_cfg:
            default_chronics_kwargs[el] = dfkwargs_cfg[el]
            
    data_feeding_kwargs = _get_default_aux(
        "data_feeding_kwargs",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=default_chronics_kwargs,
        msg_error=ERR_MSG_KWARGS["data_feeding_kwargs"],
    )
    for el in default_chronics_kwargs:
        if el not in data_feeding_kwargs:
            data_feeding_kwargs[el] = default_chronics_kwargs[el]
            
            
    ### the chronics generator
    chronics_class_used = _get_default_aux(
        "chronics_class",
        kwargs,
        defaultClassApp=GridValue,
        defaultClass=data_feeding_kwargs["chronicsClass"],
        msg_error=ERR_MSG_KWARGS["chronics_class"],
        isclass=True,
    )
    if (
        (chronics_class_used != ChangeNothing) and (chronics_class_used != FromNPY) and (chronics_class_used != FromChronix2grid)
    ) and exc_chronics is not None:
        raise EnvError(
            f"Impossible to find the chronics for your environment. Please make sure to provide "
            f'a folder "{NAME_CHRONICS_FOLDER}" within your environment folder.'
        )
    
    data_feeding_kwargs["chronicsClass"] = chronics_class_used
    if chronics_class_used.MULTI_CHRONICS:
        # add the default "gridvalueClass" in case of multi chronics and if the
        # parameters is not given in the "make" function but present in the config file
        if "gridvalueClass" not in data_feeding_kwargs:
            data_feeding_kwargs["gridvalueClass"] = grid_value_class_cfg
    
    # now build the chronics handler
    data_feeding = _get_default_aux(
        "data_feeding",
        kwargs,
        defaultClassApp=ChronicsHandler,
        defaultClass=ChronicsHandler,
        build_kwargs=data_feeding_kwargs,
        msg_error=ERR_MSG_KWARGS["chronics_handler"],
    )

    ### other rewards
    other_rewards_cfg = {}
    if "other_rewards" in config_data and config_data["other_rewards"] is not None:
        other_rewards_cfg = config_data["other_rewards"]
    other_rewards = _get_default_aux(
        "other_rewards",
        kwargs,
        defaultClassApp=dict,
        defaultinstance={},
        msg_error=ERR_MSG_KWARGS["other_rewards"],
        isclass=False,
    )
    for k in other_rewards_cfg:
        if k not in other_rewards:
            other_rewards[k] = other_rewards_cfg[k]

    # Opponent
    opponent_space_type_cfg = OpponentSpace
    if "opponent_space_type" in config_data and config_data["opponent_space_type"] is not None:
        opponent_space_type_cfg = config_data["opponent_space_type"]
    opponent_space_type = _get_default_aux(
        "opponent_space_type",
        kwargs,
        defaultClassApp=OpponentSpace,
        defaultClass=opponent_space_type_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_space_type"],
        isclass=True,
    )
    
    chronics_class_cfg = DontAct
    if (
        "opponent_action_class" in config_data
        and config_data["opponent_action_class"] is not None
    ):
        chronics_class_cfg = config_data["opponent_action_class"]
    opponent_action_class = _get_default_aux(
        "opponent_action_class",
        kwargs,
        defaultClassApp=BaseAction,
        defaultClass=chronics_class_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_action_class"],
        isclass=True,
    )
    opponent_class_cfg = BaseOpponent
    if "opponent_class" in config_data and config_data["opponent_class"] is not None:
        opponent_class_cfg = config_data["opponent_class"]
    opponent_class = _get_default_aux(
        "opponent_class",
        kwargs,
        defaultClassApp=BaseOpponent,
        defaultClass=opponent_class_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_class"],
        isclass=True,
    )
    opponent_budget_class_cfg = NeverAttackBudget
    if (
        "opponent_budget_class" in config_data
        and config_data["opponent_budget_class"] is not None
    ):
        opponent_budget_class_cfg = config_data["opponent_budget_class"]
    opponent_budget_class = _get_default_aux(
        "opponent_budget_class",
        kwargs,
        defaultClassApp=BaseActionBudget,
        defaultClass=opponent_budget_class_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_budget_class"],
        isclass=True,
    )
    opponent_init_budget_cfg = 0.0
    if (
        "opponent_init_budget" in config_data
        and config_data["opponent_init_budget"] is not None
    ):
        opponent_init_budget_cfg = config_data["opponent_init_budget"]
    opponent_init_budget = _get_default_aux(
        "opponent_init_budget",
        kwargs,
        defaultClassApp=float,
        defaultinstance=opponent_init_budget_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_init_budget"],
        isclass=False,
    )
    opponent_budget_per_ts_cfg = 0.0
    if (
        "opponent_budget_per_ts" in config_data
        and config_data["opponent_budget_per_ts"] is not None
    ):
        opponent_budget_per_ts_cfg = config_data["opponent_budget_per_ts"]
    opponent_budget_per_ts = _get_default_aux(
        "opponent_budget_per_ts",
        kwargs,
        defaultClassApp=float,
        defaultinstance=opponent_budget_per_ts_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_budget_per_ts"],
        isclass=False,
    )
    opponent_attack_duration_cfg = 0
    if (
        "opponent_attack_duration" in config_data
        and config_data["opponent_attack_duration"] is not None
    ):
        opponent_attack_duration_cfg = config_data["opponent_attack_duration"]
    opponent_attack_duration = _get_default_aux(
        "opponent_attack_duration",
        kwargs,
        defaultClassApp=int,
        defaultinstance=opponent_attack_duration_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_attack_duration"],
        isclass=False,
    )
    opponent_attack_cooldown_cfg = 99999
    if (
        "opponent_attack_cooldown" in config_data
        and config_data["opponent_attack_cooldown"] is not None
    ):
        opponent_attack_cooldown_cfg = config_data["opponent_attack_cooldown"]
    opponent_attack_cooldown = _get_default_aux(
        "opponent_attack_cooldown",
        kwargs,
        defaultClassApp=int,
        defaultinstance=opponent_attack_cooldown_cfg,
        msg_error=ERR_MSG_KWARGS["opponent_attack_cooldown"],
        isclass=False,
    )
    kwargs_opponent_cfg = {}
    if "kwargs_opponent" in config_data and config_data["kwargs_opponent"] is not None:
        kwargs_opponent_cfg = config_data["kwargs_opponent"]
    kwargs_opponent = _get_default_aux(
        "kwargs_opponent",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=kwargs_opponent_cfg,
        msg_error=ERR_MSG_KWARGS["kwargs_opponent"],
        isclass=False,
    )

    # attention budget
    has_attention_budget_cfg = False
    if (
        "has_attention_budget" in config_data
        and config_data["has_attention_budget"] is not None
    ):
        has_attention_budget_cfg = config_data["has_attention_budget"]
    has_attention_budget = _get_default_aux(
        "has_attention_budget",
        kwargs,
        defaultClassApp=bool,
        defaultinstance=has_attention_budget_cfg,
        msg_error=ERR_MSG_KWARGS["has_attention_budget"],
        isclass=False,
    )
    attention_budget_class_cfg = LinearAttentionBudget
    if (
        "attention_budget_class" in config_data
        and config_data["attention_budget_class"] is not None
    ):
        attention_budget_class_cfg = config_data["attention_budget_class"]
    attention_budget_class = _get_default_aux(
        "attention_budget_class",
        kwargs,
        defaultClassApp=LinearAttentionBudget,
        defaultClass=attention_budget_class_cfg,
        msg_error=ERR_MSG_KWARGS["attention_budget_class"],
        isclass=True,
    )

    kwargs_attention_budget_cfg = {}
    if (
        "kwargs_attention_budget" in config_data
        and config_data["kwargs_attention_budget"] is not None
    ):
        kwargs_attention_budget_cfg = config_data["kwargs_attention_budget"]
    kwargs_attention_budget = _get_default_aux(
        "kwargs_attention_budget",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=kwargs_attention_budget_cfg,
        msg_error=ERR_MSG_KWARGS["kwargs_attention_budget"],
        isclass=False,
    )

    if experimental_read_from_local_dir:
        sys_path = os.path.join(os.path.split(grid_path_abs)[0], "_grid2op_classes")
        if not os.path.exists(sys_path):
            raise RuntimeError(
                "Attempting to load the grid classes from the env path. Yet the directory "
                "where they should be placed does not exists. Did you call `env.generate_classes()` "
                "BEFORE creating an environment with `experimental_read_from_local_dir=True` ?"
            )
        if not os.path.isdir(sys_path) or not os.path.exists(
            os.path.join(sys_path, "__init__.py")
        ):
            raise RuntimeError(
                f"Impossible to load the classes from the env path. There is something that is "
                f"not a directory and that is called `_grid2op_classes`. "
                f'Please remove "{sys_path}" and call `env.generate_classes()` where env is an '
                f"environment created with `experimental_read_from_local_dir=False` (default)"
            )

    # observation key word arguments
    kwargs_observation = _get_default_aux(
        "kwargs_observation",
        kwargs,
        defaultClassApp=dict,
        defaultinstance={},
        msg_error=ERR_MSG_KWARGS["kwargs_observation"],
        isclass=False,
    )
    
    # backend for the observation 
    observation_backend_class_cfg = Backend
    if (
        "observation_backend_class" in config_data
        and config_data["observation_backend_class"] is not None
    ):
        observation_backend_class_cfg = config_data["observation_backend_class"]
    observation_backend_class = _get_default_aux(
        "observation_backend_class",
        kwargs,
        defaultClass=observation_backend_class_cfg,
        defaultClassApp=Backend,
        msg_error=ERR_MSG_KWARGS["observation_backend_class"],
        isclass=True,
    )
    if observation_backend_class is Backend:
        # in this case nothing is provided neither in the call to "make" 
        # nor in the config
        observation_backend_class = None
    
    # kwargs for observation backend
    observation_backend_kwargs_cfg_ = {"null": True} 
    # None and {} have specific meanings, so I "hack" it
    # to make the difference between "observation_backend_kwargs is not in config nor in 
    # the kwargs" and "observation_backend_kwargs is {} in the config or in the kwargs"
    observation_backend_kwargs_cfg = observation_backend_kwargs_cfg_
    if (
        "observation_backend_kwargs" in config_data
        and config_data["observation_backend_kwargs"] is not None
    ):
        observation_backend_kwargs_cfg = config_data["observation_backend_kwargs"]
    observation_backend_kwargs = _get_default_aux(
        "observation_backend_kwargs",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=observation_backend_kwargs_cfg,
        msg_error=ERR_MSG_KWARGS["kwargs_observation"],
        isclass=False,
    ) 
    if observation_backend_kwargs is observation_backend_kwargs_cfg_:
        observation_backend_kwargs = None
        
    # Finally instantiate env from config & overrides
    env = Environment(
        init_env_path=os.path.abspath(dataset_path),
        init_grid_path=grid_path_abs,
        chronics_handler=data_feeding,
        backend=backend,
        parameters=param,
        name=name_env + _add_to_name,
        names_chronics_to_backend=names_chronics_to_backend,
        actionClass=action_class,
        observationClass=observation_class,
        rewardClass=reward_class,
        legalActClass=gamerules_class,
        voltagecontrolerClass=volagecontroler_class,
        other_rewards=other_rewards,
        opponent_space_type=opponent_space_type,
        opponent_action_class=opponent_action_class,
        opponent_class=opponent_class,
        opponent_init_budget=opponent_init_budget,
        opponent_attack_duration=opponent_attack_duration,
        opponent_attack_cooldown=opponent_attack_cooldown,
        opponent_budget_per_ts=opponent_budget_per_ts,
        opponent_budget_class=opponent_budget_class,
        kwargs_opponent=kwargs_opponent,
        has_attention_budget=has_attention_budget,
        attention_budget_cls=attention_budget_class,
        kwargs_attention_budget=kwargs_attention_budget,
        logger=logger,
        _compat_glop_version=_compat_glop_version,
        _read_from_local_dir=experimental_read_from_local_dir,
        kwargs_observation=kwargs_observation,
        observation_bk_class=observation_backend_class,
        observation_bk_kwargs=observation_backend_kwargs,
    )

    # Update the thermal limit if any
    if thermal_limits is not None:
        env.set_thermal_limit(thermal_limits)

    # Set graph layout if not None and not an empty dict
    if graph_layout is not None and graph_layout:
        env.attach_layout(graph_layout)

    return env



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import os
import numpy as np
from collections import OrderedDict

from grid2op.Action import BaseAction
from grid2op.Converter.Converters import Converter
from grid2op.Exceptions.Grid2OpException import Grid2OpException
from grid2op.dtypes import dt_float, dt_int, int_types


class IdToAct(Converter):
    """
    This type of converter allows to represent action with unique id. Instead of manipulating complex objects, it allows
    to manipulate only positive integer.

    The list of all actions can either be the list of all possible unary actions (see below for a complete
    description) or by a given pre computed list.

    A "unary action" is an action that consists only in acting on one "concept" it includes:

    - disconnecting a single powerline
    - reconnecting a single powerline and connect it to bus xxx on its origin end and yyy on its extremity end
    - changing the topology of a single substation
    - performing redispatching on a single generator
    - performing curtailment on a single generator
    - performing action on a single storage unit

    Examples of non unary actions include:
    - disconnection / reconnection of 2 or more powerlines
    - change of the configuration of 2 or more substations
    - disconnection / reconnection of a single powerline and change of the configration of a single substation

    **NB** All the actions created automatically are unary. For the L2RPN 2019, agent could be allowed to act with non
    unary actions, for example by disconnecting a powerline and reconfiguring a substation. This class would not
    allow to do such action at one time step.

    **NB** The actions that are initialized by default uses the "set" way and not the "change" way (see the description
    of :class:`grid2op.BaseAction.BaseAction` for more information).

    For each powerline, 5 different actions will be computed:

    - disconnect it
    - reconnect it and connect it to bus 1 on "origin" end ann bus 1 on "extremity" end
    - reconnect it and connect it to bus 1 on "origin" end ann bus 2 on "extremity" end
    - reconnect it and connect it to bus 2 on "origin" end ann bus 1 on "extremity" end
    - reconnect it and connect it to bus 2 on "origin" end ann bus 2 on "extremity" end

    Actions corresponding to all topologies are also used by default. See
    :func:`grid2op.BaseAction.ActionSpace.get_all_unitary_topologies_set` for more information.

    In this converter:

    - `encoded_act` are positive integer, representing the index of the actions.
    - `transformed_obs` are regular observations.

    **NB** The number of actions in this converter can be especially big. For example, if a substation counts N elements
    there are roughly 2^(N-1) possible actions in this substation. This means if there are a single substation with
    more than N = 15 or 16 elements, the amount of actions (for this substation alone) will be higher than 16.000
    which makes it rather difficult to handle for most machine learning algorithm. Be carefull with that !

    """

    def __init__(self, action_space):
        Converter.__init__(self, action_space)
        self.__class__ = IdToAct.init_grid(action_space)
        self.all_actions = []
        # add the do nothing topology
        self.all_actions.append(super().__call__())
        self.n = 1
        self._init_size = action_space.size()
        self.kwargs_init = {}

    def init_converter(self, all_actions=None, **kwargs):
        """
        This function is used to initialized the converter. When the converter is created, this method should be called
        otherwise the converter might be in an unstable state.

        Parameters
        ----------
        all_actions: ``None``, ``list``, ``str``, ``np.ndarray``
            See the example section for more informations.

            if `all_actions` is:

                - ``None``: the action space will be built from scratch using the provided key word arguments.
                - a ``list``: The (ordered) list of all actions that the agent will be able to perform.
                  If given a number ``i`` the
                  converter will return action ``all_actions[i]``. In the "pacman" game, this vector could be
                  ["up", "down", "left", "right"], in this case "up" would be encode by 0, "down" by 1, "left" by 2 and
                  "right" by 3. If nothing is provided, the converter will output all the unary actions possible for
                  the environment. Be careful, computing all these actions might take some time.
                - a ``str`` this will be considered as a path where a previous converter has been saved. You need to
                  provide the full path, including the filename and its extension. It gives something like:
                  "/path/where/it/is/saved/action_space_vect.npy"

        kwargs:
            other keyword arguments (all considered to be ``True`` by default) that can be:

            set_line_status: ``bool``
                Whether you want to include the set line status in your action
                (in case the original action space allows it)

            change_line_status: ``bool``
                Whether you want to include the "change line status" in your action space
                (in case the original action space allows it)

            change_line_status: ``bool``
                Whether you want to include the "change line status" in your action space
                (in case the original action space allows it)

            set_topo_vect: ``bool``
                Whether you want to include the "set_bus" in your action space
                (in case the original action space allows it)

            change_bus_vect: ``bool``
                Whether you want to include the "change_bus" in your action space
                (in case the original action space allows it)

            redispatch: ``bool``
                Whether you want to include the "redispatch" in your action space
                (in case the original action space allows it)

            curtail: ``bool``
                Whether you want to include the "curtailment" in your action space
                (in case the original action space allows it)

            storage: ``bool``
                Whether you want to include the "storage unit" in your action space
                (in case the original action space allows it)

        Examples
        --------
        Here is an example of a code that will: make a converter by selecting some action. Save it, and then restore
        its original state to be used elsewhere.

        .. code-block:: python

            import grid2op
            from grid2op.Converter import IdToAct
            env = grid2op.make()
            converter = IdToAct(env.action_space)

            # the path were will save it
            path_ = "/path/where/it/is/saved/"
            name_file = "tmp_convert.npy"

            # init the converter, the first time, here by passing some key word arguments, to not consider
            # redispatching for example
            converter.init_converter(redispatch=False)
            converter.save(path_, name_file)

            # i just do an action, for example the number 27... whatever it does does not matter here
            act = converter.convert_act(27)

            converter2 = IdToAct(self.env.action_space)
            converter2.init_converter(all_actions=os.path.join(path_, name_file))
            act2 = converter2.convert_act(27)

            assert act ==  act2  # this is ``True`` the converter has properly been saved.


        """
        self.kwargs_init = kwargs
        if all_actions is None:
            self.all_actions = []
            # add the do nothing action, always
            self.all_actions.append(super().__call__())
            if "_set_line_status" in self._template_act.attr_list_vect:
                # lines 'set'
                include_ = True
                if "set_line_status" in kwargs:
                    include_ = kwargs["set_line_status"]
                if include_:
                    self.all_actions += self.get_all_unitary_line_set(self)

            if "_switch_line_status" in self._template_act.attr_list_vect:
                # lines 'change'
                include_ = True
                if "change_line_status" in kwargs:
                    include_ = kwargs["change_line_status"]
                if include_:
                    self.all_actions += self.get_all_unitary_line_change(self)

            if "_set_topo_vect" in self._template_act.attr_list_vect:
                # topologies 'set'
                include_ = True
                if "set_topo_vect" in kwargs:
                    include_ = kwargs["set_topo_vect"]
                if include_:
                    self.all_actions += self.get_all_unitary_topologies_set(self)

            if "_change_bus_vect" in self._template_act.attr_list_vect:
                # topologies 'change'
                include_ = True
                if "change_bus_vect" in kwargs:
                    include_ = kwargs["change_bus_vect"]
                if include_:
                    self.all_actions += self.get_all_unitary_topologies_change(self)

            if "_redispatch" in self._template_act.attr_list_vect:
                # redispatch (transformed to discrete variables)
                include_ = True
                if "redispatch" in kwargs:
                    include_ = kwargs["redispatch"]
                if include_:
                    self.all_actions += self.get_all_unitary_redispatch(self)

            if "_curtail" in self._template_act.attr_list_vect:
                # redispatch (transformed to discrete variables)
                include_ = True
                if "curtail" in kwargs:
                    include_ = kwargs["curtail"]
                if include_:
                    self.all_actions += self.get_all_unitary_curtail(self)

            if "_storage_power" in self._template_act.attr_list_vect:
                # redispatch (transformed to discrete variables)
                include_ = True
                if "storage" in kwargs:
                    include_ = kwargs["storage"]
                if include_:
                    self.all_actions += self.get_all_unitary_storage(self)

        elif isinstance(all_actions, str):
            # load the path from the path provided
            if not os.path.exists(all_actions):
                raise FileNotFoundError(
                    'No file located at "{}" where the actions should have been stored.'
                    "".format(all_actions)
                )
            try:
                all_act = np.load(all_actions)
            except Exception as e:
                raise RuntimeError(
                    'Impossible to load the data located at "{}" with error\n{}.'
                    "".format(all_actions, e)
                )
            try:
                self.all_actions = np.array([self.__call__() for _ in all_act])
                for i, el in enumerate(all_act):
                    self.all_actions[i].from_vect(el)
            except Exception as e:
                raise RuntimeError(
                    'Impossible to convert the data located at "{}" into valid grid2op action. '
                    "The error was:\n{}".format(all_actions, e)
                )
        elif isinstance(all_actions, (list, np.ndarray)):
            # assign the action to my actions
            possible_act = all_actions[0]
            if isinstance(possible_act, BaseAction):
                # list of grid2op action
                self.all_actions = np.array(all_actions)
            elif isinstance(possible_act, dict):
                # list of dictionnary (obtained with `act.as_serializable_dict()`)
                self.all_actions = np.array([self.__call__(el) for el in all_actions])
            else:
                # should be an array !
                try:
                    self.all_actions = np.array([self.__call__() for _ in all_actions])
                    for i, el in enumerate(all_actions):
                        self.all_actions[i].from_vect(el)
                except Exception as exc_:
                    raise Grid2OpException(
                        'Impossible to convert the data provided in "all_actions" into valid '
                        "grid2op action. The error was:\n{}".format(e)
                    ) from exc_
        else:
            raise RuntimeError("Impossible to load the action provided.")
        self.n = len(self.all_actions)

    def filter_action(self, filtering_fun):
        """
        This function allows you to "easily" filter generated actions.

        **NB** the action space will change after a call to this function, especially its size. It is NOT recommended
        to apply it once training has started.

        Parameters
        ----------
        filtering_fun: ``function``
            This takes an action as input and should retrieve ``True`` meaning "this action will be kept" or
            ``False`` meaning "this action will be dropped.

        """
        self.all_actions = np.array(
            [el for el in self.all_actions if filtering_fun(el)]
        )
        self.n = len(self.all_actions)

    def save(self, path, name="action_space_vect.npy"):
        """
        Save the action space as a numpy array that can be reloaded afterwards with the :func:`IdToAct.init_converter`
        function by setting argument `all_actions` to `os.path.join(path, name)`

        The resulting object will be a numpy array of float. Each row of this array will be an action of the
        action space.

        Parameters
        ----------
        path: ``str``
            The path were to save the action space

        name: ``str``, optional
            The name of the numpy array stored on disk. By default its "action_space_vect.npy"

        Examples
        --------
        Here is an example of a code that will: make a converter by selecting some action. Save it, and then restore
        its original state to be used elsewhere.

        .. code-block:: python

            import grid2op
            from grid2op.Converter import IdToAct
            env = grid2op.make()
            converter = IdToAct(env.action_space)

            # the path were will save it
            path_ = "/path/where/it/is/saved/"
            name_file = "tmp_convert.npy"

            # init the converter, the first time, here by passing some key word arguments, to not consider
            # redispatching for example
            converter.init_converter(redispatch=False)
            converter.save(path_, name_file)

            # i just do an action, for example the number 27... whatever it does does not matter here
            act = converter.convert_act(27)

            converter2 = IdToAct(self.env.action_space)
            converter2.init_converter(all_actions=os.path.join(path_, name_file))
            act2 = converter2.convert_act(27)

            assert act ==  act2  # this is ``True`` the converter has properly been saved.

        """
        if not os.path.exists(path):
            raise FileNotFoundError(
                'Impossible to save the action space as the directory "{}" does not exist.'
                "".format(path)
            )
        if not os.path.isdir(path):
            raise NotADirectoryError(
                'The path to save the action space provided "{}" is not a directory.'
                "".format(path)
            )
        saved_npy = (
            np.array([el.to_vect() for el in self.all_actions])
            .astype(dtype=dt_float)
            .reshape(self.n, -1)
        )
        np.save(file=os.path.join(path, name), arr=saved_npy)

    def sample(self):
        """
        Having define a complete set of observation an agent can do, sampling from it is now made easy.

        One action amoung the n possible actions is used at random.

        Returns
        -------
        res: ``int``
            An id of an action.

        """
        idx = self.space_prng.randint(0, self.n, dtype=dt_int)
        return idx

    def convert_act(self, encoded_act):
        """
        In this converter, we suppose that "encoded_act" is an id of an action stored in the
        :attr:`IdToAct.all_actions` list.

        Converting an id of an action (here called "act") into a valid action is then easy: we just need to take the
        "act"-th element of :attr:`IdToAct.all_actions`.

        Parameters
        ----------
        encoded_act: ``int``
            The id of the action

        Returns
        -------
        action: :class:`grid2op.Action.Action`
            The action corresponding to id "act"
        """

        return self.all_actions[encoded_act]

    def get_gym_dict(self, cls_gym):
        """
        Transform this converter into a dictionary that can be used to initialized a :class:`gym.spaces.Dict`.
        The converter is modeled as a "Discrete" gym space with as many elements as the number
        of different actions handled by this converter.

        This is available as the "action" keys of the spaces.Dict gym action space build from it.

        This function should not be used "as is", but rather through :class:`grid2op.Converter.GymConverter`

        cls_gym represents either :class:`grid2op.gym_compat.LegacyGymActionSpace` or
        :class:`grid2op.gym_compat.GymnasiumActionSpace`
        """
        res = {"action": cls_gym._DiscreteType(n=self.n)}
        return res

    def convert_action_from_gym(self, gymlike_action):
        """
        Convert the action (represented as a gym object, in fact an ordered dict) as an action
        compatible with this converter.

        This is not compatible with all converters and you need to install gym for it to work.

        Parameters
        ----------
        gymlike_action:
            the action to be converted to an action compatible with the action space representation

        Returns
        -------
        res:
            The action converted to be understandable by this converter.

        Examples
        ---------
        Here is an example on how to use this feature with the :class:`grid2op.Converter.IdToAct`
        converter (imports are not shown here).

        .. code-block:: python

            # create the environment
            env = grid2op.make()

            # create the converter
            converter = IdToAct(env.action_space)

            # create the gym action space
            gym_action_space = GymObservationSpace(action_space=converter)

            gym_action = gym_action_space.sample()
            converter_action = converter.from_gym(gym_action)  # this represents the same action
            grid2op_action = converter.convert_act(converter_action)  # this is a grid2op action

        """
        res = gymlike_action["action"]
        if not isinstance(res, int_types):
            raise RuntimeError("TODO")
        return int(res)

    def convert_action_to_gym(self, action):
        """
        Convert the action (compatible with this converter) into a "gym action" (ie an OrderedDict)

        This is not compatible with all converters and you need to install gym for it to work.

        Parameters
        ----------
        action:
            the action to be converted to an action compatible with the action space representation

        Returns
        -------
        res:
            The action converted to a "gym" model (can be used by a machine learning model)

        Examples
        ---------
        Here is an example on how to use this feature with the :class:`grid2op.Converter.IdToAct`
        converter (imports are not shown here).

        .. code-block:: python

            # create the environment
            env = grid2op.make()

            # create the converter
            converter = IdToAct(env.action_space)

            # create the gym action space
            gym_action_space = GymObservationSpace(action_space=converter)

            converter_action = converter.sample()
            gym_action = converter.to_gym(converter_action)  # this represents the same action

        """
        res = OrderedDict({"action": int(action)})
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import numpy as np
from grid2op.Opponent.BaseActionBudget import BaseActionBudget


class NeverAttackBudget(BaseActionBudget):
    """
    This class define an unlimited budget for the opponent.

    It SHOULD NOT be used if the opponent is allowed to take any actions!
    """

    def __init__(self, action_space):
        BaseActionBudget.__init__(self, action_space)

    def __call__(self, attack):
        if attack is not None:
            return np.inf
        return 0.0



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import sys
from tqdm import tqdm
import re

import tarfile

from grid2op.Exceptions import Grid2OpException

try:
    import urllib.request
except Exception as e:
    raise RuntimeError("Impossible to find library urllib. Please install it.")

URL_GRID2OP_DATA = "https://github.com/Tezirg/Grid2Op/releases/download/{}/{}"
DATASET_TAG_v0_1_0 = "datasets-v0.1.0"
DICT_URL_GRID2OP_DL = {
    "rte_case14_realistic": URL_GRID2OP_DATA.format(
        DATASET_TAG_v0_1_0, "rte_case14_realistic.tar.bz2"
    ),
    "rte_case14_redisp": URL_GRID2OP_DATA.format(
        DATASET_TAG_v0_1_0, "rte_case14_redisp.tar.bz2"
    ),
    "l2rpn_2019": URL_GRID2OP_DATA.format(DATASET_TAG_v0_1_0, "l2rpn_2019.tar.bz2"),
}
LI_VALID_ENV = sorted(['"{}"'.format(el) for el in DICT_URL_GRID2OP_DL.keys()])


class DownloadProgressBar(tqdm):
    """
    INTERNAL

     .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    This class is here to show the progress bar when downloading this dataset
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    """
    INTERNAL

     .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    This function download the file located at 'url' and save it to 'output_path'

    Parameters
    ----------
    url: ``str``
        The url of the file to download

    output_path: ``str``
        The path where the data will be stored.
    """
    with DownloadProgressBar(
        unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
    ) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def _aux_download(url, dataset_name, path_data, ds_name_dl=None):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
    """
    if ds_name_dl is None:
        ds_name_dl = dataset_name
    final_path = os.path.join(path_data, ds_name_dl)
    if os.path.exists(final_path):
        str_ = (
            'Downloading and extracting this data would create a folder "{final_path}" '
            "but this folder already exists. Either you already downloaded the data, in this case "
            "you can invoke the environment from a python script with:\n"
            '\t env = grid2op.make("{final_path}")\n'
            'Alternatively you can also delete the folder "{final_path}" from your computer and run this command '
            "again.\n"
            "Finally, you can download the data in a different folder by specifying (in a command prompt):\n"
            '\t grid2op.download --name "{env_name}" --path_save PATH\WHERE\YOU\WANT\TO\DOWNLOAD'
            "".format(final_path=final_path, env_name=dataset_name)
        )
        print(str_)
        raise Grid2OpException(str_)

    if not os.path.exists(path_data):
        print(
            'Creating path "{}" where data for "{}" environment will be downloaded.'
            "".format(path_data, ds_name_dl)
        )
        try:
            os.mkdir(path_data)
        except Exception as exc_:
            str_ = (
                'Impossible to create path "{}" to store the data. Please save the data in a different repository '
                'with setting the argument "--path_save"'
                "Error was:\n{}".format(path_data, exc_)
            )
            raise Grid2OpException(str_)

    output_path = os.path.abspath(
        os.path.join(path_data, "{}.tar.bz2".format(ds_name_dl))
    )

    # download the data (with progress bar)
    print("downloading the training data, this may take a while.")
    download_url(url, output_path)

    tar = tarfile.open(output_path, "r:bz2")
    print('Extract the tar archive in "{}"'.format(os.path.abspath(path_data)))
    tar.extractall(path_data)
    tar.close()

    # rename the file if necessary
    if ds_name_dl != dataset_name:
        try:
            os.rename(final_path, os.path.join(path_data, dataset_name))
        except FileNotFoundError as exc_:
            # the try catch is added because for some environments, the
            # archive name does not match the env name.
            # so the folder cannot be deleted properly.
            print(f"WARN: file \"{final_path}\" could not be found. You might need to manually delete it.")
            pass

    # and rm the tar bz2
    # bug in the AWS file... named ".tar.tar.bz2" ...
    os.remove(output_path)

    # check for update (if any)
    from grid2op.MakeEnv.UpdateEnv import _update_files

    _update_files(dataset_name)

    print(
        'You may now use the environment "{}" with the available data by invoking:\n'
        '\tenv = grid2op.make("{}")'
        "".format(dataset_name, dataset_name)
    )


def main_download(dataset_name, path_data):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    """
    dataset_name = dataset_name.lower().rstrip().lstrip()
    dataset_name = re.sub('"', "", dataset_name)

    if dataset_name not in DICT_URL_GRID2OP_DL:
        print(
            'Impossible to find environment named "{env_name}". Known environments are:\n{li_env}'
            "".format(env_name=dataset_name, li_env=",".join(LI_VALID_ENV))
        )
        sys.exit(1)

    url = DICT_URL_GRID2OP_DL[dataset_name]
    _aux_download(url, dataset_name, path_data)
